# -*- coding: utf-8 -*-

'''
Tests for ReCalc that do not involve tkinter.

testing_ReCalc.py
'''

import unittest
import doctest
import os
import math
import string

import numpy as np

class NonRepeatingList(object):
	'''
	A mutable list that doesn't have two of the same element in a row.

	>>> repr(NonRepeatingList(3, 3, 4))
	'NonRepeatingList(*[3, 4])'
	'''

	def __init__(self, *args):
		if len(args) > 0:
			self.items = [args[0]]
			for i in args:
				if i != self.items[-1]:
					self.items.append(i)
		else:
			self.items = []

	def __getitem__(self, index):
		return(self.items[index])

	def __delitem__(self, index):
		del self.items[index]
		if index != 0:
			if self.items[index] == self.items[index - 1]:
				del self.items[index]

	def __contains__(self, item):
		return(item in self.items)

	def __len__(self):
		return(len(self.items))

	def __repr__(self):
		return("NonRepeatingList(*" + repr(self.items) + ")")

	def __str__(self):
		return(str(self.items))

	def __eq__(self, other):
		if isinstance(other, NonRepeatingList):
			if self.items == other.items:
				return(True)
		return(False)

	def append(self, *args):
		for item in args:
			if len(self.items) > 0:
				if self.items[-1] != item:
					self.items.append(item)
			else:
				self.items.append(item)

	def clear(self):
		self.items.clear()


import ReCalc as c


class TestCheckIfFloat(unittest.TestCase):
	'''
	Test that the check_if_float function works as expected.
	'''

	def test_float(self):
		'''
		Check that it works for floats.
		'''

		self.assertTrue(c.check_if_float(3.4))

	def test_int_type(self):
		'''
		Check that it works for ints.
		'''

		self.assertTrue(c.check_if_float(4))

	def test_string_type(self):
		'''
		Check that it works for floats as strings.
		'''

		self.assertTrue(c.check_if_float("3.4"))

	def test_word(self):
		'''
		Check that it returns false for words.
		'''

		self.assertFalse(c.check_if_float("hello"))

	def test_int_string(self):
		'''
		Check it works for ints as strings.
		'''

		self.assertTrue(c.check_if_float("2"))

	def test_mixed(self):
		self.assertFalse(c.check_if_float("sin(3.14)"))

	def test_parrentheses(self):
		self.assertFalse(c.check_if_float("(5)"))

	def test_list_type(self):
		self.assertFalse(c.check_if_float([2]))


class Files(unittest.TestCase):
	'''
	Check that require files and paths are present and exist.
	'''

	def test_main_path(self):
		'''
		Test that the path of the module actually exists.
		'''

		self.assertTrue(os.path.exists(c.calc_path))

	def test_picked_data(self):
		'''
		Check that the information file is present.
		'''

		self.assertTrue(
			os.path.isfile(
				os.path.join(c.calc_path, "ReCalc_info.txt")))

	def test_tkinter_icon(self):
		'''
		Check that the icon for the window is present.
		'''

		self.assertTrue(
			os.path.isfile(
				os.path.join(c.calc_path, "ReCalc_icon.ico")))


class SwitchDegreeMode(unittest.TestCase):
	'''
	Test the switch_degree_mode function works as expected.
	'''

	def setUp(self):
		'''
		Remember what the degree_mode started as.
		'''

		self.original_degree_mode = c.degree_mode

	def test_set_degree_mode(self):
		c.switch_degree_mode(2)
		self.assertEqual(c.degree_mode, 2)
		self.assertEqual(c.calc_info["degree_mode"], 2)

	def test_set_radian_mode(self):
		c.switch_degree_mode(0)
		self.assertEqual(c.degree_mode, 0)
		self.assertEqual(c.calc_info["degree_mode"], 0)

	def test_set_degree_mode_string(self):
		c.switch_degree_mode("degree")
		self.assertEqual(c.degree_mode, 2)
		self.assertEqual(c.calc_info["degree_mode"], 2)

	def test_set_radian_mode_string(self):
		c.switch_degree_mode("radian")
		self.assertEqual(c.degree_mode, 0)
		self.assertEqual(c.calc_info["degree_mode"], 0)

	def test_set_not_option(self):
		'''
		Check that switch_degree_mode throws an error if you pass it
		an invalid integer.
		'''

		with self.assertRaises(c.CalculatorError):
			c.switch_degree_mode(1)

	def test_set_to_random_word(self):
		'''
		Check that switch_degree_mode throws an error if you pass it
		an invalid string.
		'''

		with self.assertRaises(c.CalculatorError):
			c.switch_degree_mode("random")

	def tearDown(self):
		'''Set the degree_mode back to what it started as.'''

		c.switch_degree_mode(self.original_degree_mode)


class SwitchPolarCartesian(unittest.TestCase):
	'''
	Test switching between polar and Cartesian graphing modes works.
	'''

	def setUp(self):
		'''
		Remember what mode it started in.
		'''

		self.original_graphing_mode = c.polar_mode

	def test_set_polar(self):
		c.switch_polar_mode("polar")
		self.assertEqual(c.polar_mode, True)
		self.assertEqual(c.calc_info["polar_mode"], True)

	def test_set_cartesian(self):
		c.switch_polar_mode("Cartesian")
		self.assertEqual(c.polar_mode, False)
		self.assertEqual(c.calc_info["polar_mode"], False)

	def test_set_polar_boolean(self):
		c.switch_polar_mode(True)
		self.assertEqual(c.polar_mode, True)
		self.assertEqual(c.calc_info["polar_mode"], True)

	def test_set_cartesian_boolean(self):
		c.switch_polar_mode(False)
		self.assertEqual(c.polar_mode, False)
		self.assertEqual(c.calc_info["polar_mode"], False)

	def test_set_not_option(self):
		with self.assertRaises(c.CalculatorError):
			c.switch_polar_mode(4)

	def tearDown(self):
		c.switch_polar_mode(self.original_graphing_mode)


class FindMatch(unittest.TestCase):
	'''
	Test the find match function.
	'''

	def test_basic_functionality(self):
		self.assertEqual(
			c.find_match("(inside)outside"), ('(inside)', 'outside'))

	def test_extra_parentheses(self):
		self.assertEqual(
			c.find_match("((inside)))"), ('((inside))', ')'))

	def test_mismatched_parentheses_left(self):
		with self.assertRaises(c.CalculatorError):
			c.find_match("(word")

	def test_mismatched_parentheses_right(self):
		with self.assertRaises(c.CalculatorError):
			c.find_match("word)")


class BracketsFunction(unittest.TestCase):
	'''
	Test the brackets function.
	'''

	def test_matched_brackets(self):
		self.assertTrue(c.brackets(""))
		self.assertTrue(c.brackets("()"))
		self.assertTrue(c.brackets("(())"))
		self.assertTrue(c.brackets("(()())"))
		self.assertTrue(c.brackets("((word)(other)word)"))

	def test_unmatched_brackets(self):
		self.assertFalse(c.brackets("("))
		self.assertFalse(c.brackets(")"))
		self.assertFalse(c.brackets(")("))
		self.assertFalse(c.brackets("(()words"))


class Separate(unittest.TestCase):
	'''
	Test that the separate function works as expected.
	'''

	def test_split_no_parentheses(self):
		self.assertEqual(c.separate("2,4,5,6"), ("2", "4", "5", "6"))

	def test_split_parentheses(self):
		self.assertEqual(
			c.separate("3, 4, (5,6), 3"), ("3", " 4", " (5,6)", " 3"))

	def test_split_nested_parentheses(self):
		self.assertEqual(
			c.separate("1, (2, 3, (4, 5), 6), 7, (8), (9, 10)"),
			("1", " (2, 3, (4, 5), 6)", " 7", " (8)", " (9, 10)"))


class Constant(unittest.TestCase):
	'''
	Test that the constant function works as exprected.
	'''

	def test_all_the_constants(self):
		'''
		Test all the constants in the constant function.
		'''

		self.assertEqual(c.constant("pi"), math.pi)
		self.assertEqual(c.constant("e"), math.e)
		self.assertEqual(c.constant("phi"), (1 + 5 ** 0.5) / 2)
		self.assertEqual(c.constant("π"), math.pi)
		self.assertEqual(c.constant("ans"), c.ans)
		self.assertEqual(c.constant("answer"), c.ans)
		self.assertEqual(c.constant("tau"), math.tau)
		self.assertEqual(c.constant("τ"), math.tau)
		self.assertEqual(c.constant("φ"), (1 + 5 ** 0.5) / 2)

	def test_unknown_constant(self):
		'''
		Test that the constant function returns an error string if
		it gets passed an unknown constant.
		'''

		with self.assertRaises(c.CalculatorError):
			c.constant("pie")


class SolvingEquations(unittest.TestCase):
	'''
	Test solving equations works.
	'''

	def test_solve_with_no_equals_or_variable(self):
		self.assertEqual(c.solve_equations("x-5"), 5)

	def test_solve_with_no_equals(self):
		self.assertEqual(c.solve_equations("t-5 for t"), 5)

	def test_solve_with_no_variable(self):
		self.assertEqual(c.solve_equations("2*x = 4"), 2)

	def test_solve_with_everything(self):
		self.assertEqual(c.solve_equations("2*t = 4 for t"), 2)


class EvaluationFunction(unittest.TestCase):
	'''
	Test the evaluation function.
	'''

	def test_eval_without_variable(self):
		self.assertEqual(c.evaluate("x", "5"), "5")

	def test_eval_with_variable(self):
		self.assertEqual(c.evaluate("t", "5", var = "t"), "5")

	@unittest.expectedFailure
	def test_double_eval_function(self):
		'''
		Test nested evaluate statements.
		'''

		self.assertEqual(
			c.evaluate("eval t at x for t", "4"), "4")


class FindDerivative(unittest.TestCase):
	'''
	Test the find_derivative function.
	'''

	def test_find_basic_derivative(self):
		self.assertAlmostEqual(
			float(c.find_derivative("x^2", "3")), 6.0)

	def test_find_derivative_with_variable(self):
		self.assertAlmostEqual(
			float(c.find_derivative("t^2", "3", var = "t")), 6.0)

	@unittest.expectedFailure
	def test_double_derivative(self):
		'''
		Test nested derivative statements.
		'''

		self.assertAlmostEqual(
			float(c.find_derivative("x * derivative of t^2 at 3", "1")),
			6.0)


class IntegrateFunction(unittest.TestCase):
	'''
	Check that integration works as expected.
	'''

	def test_basic_integral(self):
		self.assertEqual(c.integrate_function("2*x", "x", "0", "3"), "9")

	def test_other_variables(self):
		'''
		Test integration works with respect to any lowercase variable.

		It will not work with 'I', 'E', 'S', 'N', 'C', 'O', or 'Q'
		because sympy uses those letters to represent other things.
		'''

		for i in string.ascii_lowercase:
			with self.subTest(i = i):
				self.assertEqual(
					c.integrate_function("2*%s" % i, i, "0", "3"),
					"9",
					msg = "Integrating with respect to %s failed" % i)

			self.assertEqual(
				c.integrate_function("2*A", "A", "0", "3"), "9")

	def test_integrate_at_expression(self):
		'''
		Test that integrate can integrate from a point that is not
		a single number.
		'''

		self.assertEqual(
			c.integrate_function("2*x", "x", "0", "1+2"), "9")


class CombinationsAndPermutations(unittest.TestCase):
	'''
	Test that combinations_and_permutations works as expected.
	'''

	def test_basic_combination_choose_notation(self):
		self.assertEqual(
			c.combinations_and_permutations("choose", "C", "5", "2"),
			"10.0")

	def test_basic_permutaion_choose_notation(self):
		self.assertEqual(
			c.combinations_and_permutations("choose", "P", "5", "2"),
			"20.0")

	def test_basic_combination_func_notation(self):
		self.assertEqual(
			c.combinations_and_permutations("func", "C", "(5,2)"),
			"10.0")

	def test_basic_permutaion_func_notation(self):
		self.assertEqual(
			c.combinations_and_permutations("func", "P", "(5,2)"),
			"20.0")

	def test_nested_combinations(self):
		self.assertEqual(
			c.combinations_and_permutations("func", "C", "(5,C(2,1))"),
			"10.0")

	def test_nested_permutations(self):
		self.assertEqual(
			c.combinations_and_permutations("func", "P", "(5,P(2,1))"),
			"20.0")

	def test_nested_mixed(self):
		self.assertEqual(
			c.combinations_and_permutations("func", "P", "(C(5,1),P(2,1))"),
			"20.0")

	def test_nested_mixed2(self):
		self.assertEqual(
			c.combinations_and_permutations("func", "C", "(C(5,1),P(2,1))"),
			"10.0")

	def test_choose_notation_needs_four_arguments(self):
		self.assertEqual(
			c.combinations_and_permutations("choose", "C", "(5,2)"),
			("ERROR: combinations_and_permutations requires a fourth "
				"argument when using choose notation"))

	def test_function_notation_needs_only_three_arguments(self):
		self.assertEqual(
			c.combinations_and_permutations("func", "C", "5", m = "2"),
			("ERROR: combinations_and_permutations can not take a "
				"fourth argument when using function notation"))

	def test_raises_errors_for_not_c_or_p_choose_notaion(self):
		with self.assertRaises(c.CalculatorError):
			c.combinations_and_permutations("choose", "R", "5", "2")

	def test_raises_errors_for_not_c_or_p_func_notaion(self):
		with self.assertRaises(c.CalculatorError):
			c.combinations_and_permutations("func", "R", "(5, 2)")


class StatisticsFunctions(unittest.TestCase):
	'''
	Test that statistics_functions works as expected.
	'''

	def test_all_basic_statistics(self):
		'''Test basic usage of statistics functions.'''

		self.assertEqual(
			c.statistics_functions("mean", "(4, 6, 39, 2, 11)"),
			"12.4")
		self.assertEqual(
			c.statistics_functions("ave", "(4, 6, 39, 2, 11)"),
			"12.4")
		self.assertEqual(
			c.statistics_functions("mean", "(4, 6, 39, 2, 11)"),
			"12.4")
		self.assertEqual(
			c.statistics_functions("median", "(1, 2, 3, 1, 7)"),
			"2")
		self.assertEqual(
			c.statistics_functions("mode", "(1, 2, 3, 1, 7)"),
			"1")
		self.assertEqual(
			c.statistics_functions("min", "(3, 6, 11, -7, -1, 55)"),
			"-7")
		self.assertEqual(
			c.statistics_functions("max", "(3, 6, 11, -7, -1, 55)"),
			"55")
		self.assertAlmostEqual(
			float(c.statistics_functions("stdev", "(3, 4, 4, 6, 8)")),
			2.0)

	def test_nested_stats_functions(self):
		self.assertEqual(
			c.statistics_functions("mean", "(1, 2, mean(1, 3, 5))"),
			"2")

	def test_not_defined_functions(self):
		'''
		Test that statistics_functions returns an error string
		when you pass it an undefined function.
		'''

		with self.assertRaises(c.CalculatorError):
			c.statistics_functions("sin", "(pi)")


class SingleArgumentFunction(unittest.TestCase):
	'''
	Test the single_argument function.
	'''

	def test_basic_trig_functions(self):
		self.assertAlmostEqual(
			float(c.single_argument("sin", "(pi/6)")),
			0.5)
		self.assertAlmostEqual(
			float(c.single_argument("cos", "(pi/3)")),
			0.5)
		self.assertAlmostEqual(
			float(c.single_argument("tan", "(pi/4)")),
			1.0)
		self.assertAlmostEqual(
			float(c.single_argument("sec", "(pi/3)")),
			2.0)
		self.assertAlmostEqual(
			float(c.single_argument("csc", "(pi/6)")),
			2.0)
		self.assertAlmostEqual(
			float(c.single_argument("cot", "(pi/4)")),
			1.0)

	def test_hyperbolic_trig_functions(self):
		self.assertAlmostEqual(
			float(c.single_argument("sinh", "(1)")),
			(math.e ** 2 - 1) / (2 * math.e))
		self.assertAlmostEqual(
			float(c.single_argument("cosh", "(1)")),
			(math.e ** 2 + 1) / (2 * math.e))
		self.assertAlmostEqual(
			float(c.single_argument("tanh", "(1)")),
			(math.e ** 2 - 1) / (math.e ** 2 + 1))
		self.assertAlmostEqual(
			float(c.single_argument("csch", "(1)")),
			(2 * math.e) / (math.e ** 2 - 1))
		self.assertAlmostEqual(
			float(c.single_argument("sech", "(1)")),
			(2 * math.e) / (math.e ** 2 + 1))
		self.assertAlmostEqual(
			float(c.single_argument("coth", "(1)")),
			(math.e ** 2 + 1) / (math.e ** 2 - 1))

	def test_inverse_trig_functions(self):
		self.assertAlmostEqual(
			float(c.single_argument("asin", "(.5)")),
			math.pi / 6)
		self.assertAlmostEqual(
			float(c.single_argument("acos", "(.5)")),
			math.pi / 3)
		self.assertAlmostEqual(
			float(c.single_argument("atan", "(3 ** .5)")),
			math.pi / 3)
		self.assertAlmostEqual(
			float(c.single_argument("asec", "(2)")),
			math.pi / 3)
		self.assertAlmostEqual(
			float(c.single_argument("acsc", "(2)")),
			math.pi / 6)
		self.assertAlmostEqual(
			float(c.single_argument("acot", "(3 ** .5)")),
			math.pi / 6)

	def test_inverse_hyperbolic_functions(self):
		'''Test basic usage of inverse hyperbolic trig functions.'''

		self.assertAlmostEqual(
			float(c.single_argument(
				"asinh",
				"(((e ** 2) - 1) / (2 * e))")),
			1)
		self.assertAlmostEqual(
			float(c.single_argument(
				"acosh",
				"((e ** 2 + 1) / (2 * e))")),
			1)
		self.assertAlmostEqual(
			float(c.single_argument(
				"atanh",
				"((e ** 2 - 1) / (e ** 2 + 1))")),
			1)
		self.assertAlmostEqual(
			float(c.single_argument(
				"acsch",
				"((2 * e) / (e ** 2 - 1))")),
			1)
		self.assertAlmostEqual(
			float(c.single_argument(
				"asech",
				"((2 * e) / (e ** 2 + 1))")),
			1)
		self.assertAlmostEqual(
			float(c.single_argument(
				"acoth",
				"((e ** 2 + 1) / (e ** 2 - 1))")),
			1)

	def test_archaic_trig_functions(self):
		'''Test basic usage of old mostly abandoned trig functions.'''

		self.assertAlmostEqual(
			float(c.single_argument("versin", "(pi/3)")),
			.5)
		self.assertAlmostEqual(
			float(c.single_argument("vercosin", "(pi/3)")),
			1.5)
		self.assertAlmostEqual(
			float(c.single_argument("coversin", "(pi/6)")),
			.5)
		self.assertAlmostEqual(
			float(c.single_argument("covercosin", "(pi/6)")),
			1.5)
		self.assertAlmostEqual(
			float(c.single_argument("haversin", "(pi/3)")),
			.25)
		self.assertAlmostEqual(
			float(c.single_argument("havercosin", "(pi/3)")),
			.75)
		self.assertAlmostEqual(
			float(c.single_argument("hacoversin", "(pi/6)")),
			.25)
		self.assertAlmostEqual(
			float(c.single_argument("hacovercosin", "(pi/6)")),
			.75)
		self.assertAlmostEqual(
			float(c.single_argument("exsec", "(0)")),
			0)
		self.assertAlmostEqual(
			float(c.single_argument("excsc", "(pi/2)")),
			0)

	def test_other_functions(self):
		'''Test basic usage of single argument functions that
		are not in anyway trig functions.'''

		self.assertEqual(
			c.single_argument("abs", "(-1)"),
			"1")
		self.assertEqual(
			c.single_argument("abs", "(3)"),
			"3")
		self.assertEqual(
			c.single_argument("ceil", "(.5)"),
			"1")
		self.assertEqual(
			c.single_argument("ceil", "(-1.5)"),
			"-1")
		self.assertEqual(
			c.single_argument("floor", "(4.8)"),
			"4")
		self.assertEqual(
			c.single_argument("floor", "(-4.8)"),
			"-5")
		self.assertAlmostEqual(
			float(c.single_argument("erf", "(e)")),
			0.9998790689599)

	def test_trig_functions_with_degree_symbol(self):
		self.assertAlmostEqual(
			float(c.single_argument("sin", "(30°)")),
			0.5)
		self.assertAlmostEqual(
			float(c.single_argument("cos", "(60°)")),
			0.5)
		self.assertAlmostEqual(
			float(c.single_argument("tan", "(45°)")),
			1.0)
		self.assertAlmostEqual(
			float(c.single_argument("sec", "(60°)")),
			2.0)
		self.assertAlmostEqual(
			float(c.single_argument("csc", "(30°)")),
			2.0)
		self.assertAlmostEqual(
			float(c.single_argument("cot", "(45°)")),
			1.0)

	def test_nested_trig_functions(self):
		self.assertAlmostEqual(
			float(c.single_argument("sin", "(asin(.56))")),
			.56)
		self.assertAlmostEqual(
			float(c.single_argument("cos", "(sin(0))")),
			1)
		self.assertAlmostEqual(
			float(c.single_argument("sin", "(cos(pi/2))")),
			0)
		self.assertAlmostEqual(
			float(c.single_argument("sin", "(asin(.12))")),
			.12)
		self.assertAlmostEqual(
			float(c.single_argument("asin", "(sin(1.33))")),
			1.33)


class InverseFunctionsDegreeMode(unittest.TestCase):
	'''
	Test inverse hyperbolic and trig function while in degree mode.
	'''

	def setUp(self):
		'''
		Remember what mode it started in. Then switch to degree mode.
		'''

		self.mode = c.degree_mode
		c.switch_degree_mode(2)

	def test_inverse_trig_functions_degree_mode(self):
		self.assertAlmostEqual(
			float(c.single_argument("asin", "(.5)")),
			30)
		self.assertAlmostEqual(
			float(c.single_argument("acos", "(.5)")),
			60)
		self.assertAlmostEqual(
			float(c.single_argument("atan", "(1)")),
			45)
		self.assertAlmostEqual(
			float(c.single_argument("asec", "(2)")),
			60)
		self.assertAlmostEqual(
			float(c.single_argument("acsc", "(2)")),
			30)
		self.assertAlmostEqual(
			float(c.single_argument("acot", "(1)")),
			45)

	def test_inverse_hyperbolic_functions_degree_mode(self):
		self.assertAlmostEqual(
			float(c.single_argument(
				"asinh", "(0.5478534738880397)")),
			30)
		self.assertAlmostEqual(
			float(c.single_argument(
				"acosh", "(1.600286857702386)")),
			60)
		self.assertAlmostEqual(
			float(c.single_argument(
				"atanh", "(0.6557942026326724)")),
			45)
		self.assertAlmostEqual(
			float(c.single_argument(
				"asech", "(0.6248879662960872)")),
			60)
		self.assertAlmostEqual(
			float(c.single_argument(
				"acsch", "(1.8253055746879534)")),
			30)
		self.assertAlmostEqual(
			float(c.single_argument(
				"acoth", "(1.5248686188220641)")),
			45)

	def tearDown(self):
		'''
		Go back to what mode it started in.
		'''

		c.switch_degree_mode(self.mode)


class GammaAndFactorial(unittest.TestCase):
	'''
	Test the gamma and factorial functions.
	'''

	def test_gamma(self):
		for i in np.arange(1.0, 10.0, .5):
			with self.subTest(i = i):
				self.assertAlmostEqual(
					float(c.gamma("(%s)" % i)), math.gamma(i))

	def test_factorial(self):
		for i in range(10):
			with self.subTest(i = i):
				self.assertEqual(
					c.factorial(str(i)), math.gamma(i + 1))


class Logarithm(unittest.TestCase):
	'''
	Test the logarithm function.
	'''

	def test_log(self):
		'''
		Test logarithm with base 3.
		'''

		self.assertEqual(c.logarithm("(9, 3)"), "2.0")

	def test_log2(self):
		'''
		Test logarithm with base 2.
		'''

		self.assertEqual(c.logarithm("(8, 2)"), "3.0")

	def test_log10(self):
		'''
		Test logarithm with base 10.
		'''

		self.assertEqual(c.logarithm("(.1, 10)"), "-1.0")

	def test_ln(self):
		'''
		Test logarithm with one argument.
		'''

		self.assertEqual(c.logarithm("(e^2)"), "2.0")

	def test_nested_logs(self):
		'''
		Test logarithm with another log as an argument.
		'''

		self.assertEqual(c.logarithm("(27, log(8, 2))"), "3.0")


class Modulus(unittest.TestCase):
	'''
	Test the modulus function.
	'''

	def test_modulus(self):
		self.assertEqual(c.modulus("(7, 4)"), "3.0")
		self.assertEqual(c.modulus("(9, 4)"), "1.0")
		self.assertEqual(c.modulus("(9, 3)"), "0.0")
		self.assertEqual(c.modulus("(5, 2)"), "1.0")

	def test_nested_modulus(self):
		'''
		Test modulus with another modulus as an argument.
		'''

		self.assertEqual(c.modulus("(11, mod(3, 4))"), "2.0")


class AbsoluteValue(unittest.TestCase):
	'''
	Test the absolute_value function.
	'''

	def test_basic(self):
		self.assertEqual(c.abs_value("|-1|"), "1.0")

	def test_expression(self):
		self.assertEqual(c.abs_value("|4-9|"), "5.0")

	def test_starts_with_negative(self):
		self.assertEqual(c.abs_value("|-1*3|"), "3.0")

	def test_nested(self):
		self.assertEqual(c.abs_value("||5-6||"), "|1.0|")

	def test_nested_negative(self):
		self.assertEqual(c.abs_value("|-1*|2-6||"), "|-1*4.0|")

	def test_subtract_abs_val(self):
		self.assertEqual(c.abs_value("|3-|1-7||"), "|3-6.0|")

	def test_double_nested(self):
		self.assertEqual(c.abs_value("|1-|3+|-4|||"), "|1-|3+4.0||")

	def test_parts_outside_abs_value(self):
		self.assertEqual(c.abs_value("4+|-3|"), "4+3.0")

	def test_multiply_abs_value(self):
		self.assertEqual(c.abs_value("2*|3-1|"), "2*2.0")

	def test_two_separate_abs_values(self):
		self.assertEqual(c.abs_value("|3-2|+|-1*3|"), "1.0+|-1*3|")

	def test_unmatched_pipes(self):
		self.assertEqual(
			c.abs_value("|1"),
			("ERROR: There must be an even number of pipes in an "
			"absolute value expression."))


class Comma(unittest.TestCase):
	'''
	Test the comma function.
	'''

	def test_remove_commas(self):
		self.assertEqual(c.comma("456", "345"), "456345")
		self.assertEqual(c.comma("6", "345"), "6345")
		self.assertEqual(c.comma("21", "345.3333"), "21345.3333")
		self.assertEqual(c.comma("45", "345."), "45345.")

	def test_returns_error_string_if_wrong(self):
		with self.assertRaises(c.CalculatorError):
			c.comma("21", "32")
		with self.assertRaises(c.CalculatorError):
			c.comma("21", "3234")
		with self.assertRaises(c.CalculatorError):
			c.comma("21", "32.123")
		with self.assertRaises(c.CalculatorError):
			c.comma("21", "")


class Regex(unittest.TestCase):
	'''
	Test that the regular expression match the correct strings in the
	correct way.
	'''

	def test_reg_num_matches(self):
		self.assertRegex("1", "^" + c.reg_num + "$")
		self.assertRegex("2.0", "^" + c.reg_num + "$")
		self.assertRegex("4.6", "^" + c.reg_num + "$")
		self.assertRegex("30", "^" + c.reg_num + "$")
		self.assertRegex("-5", "^" + c.reg_num + "$")
		self.assertRegex("-3.0", "^" + c.reg_num + "$")
		self.assertRegex("-5.1", "^" + c.reg_num + "$")
		self.assertRegex("-54", "^" + c.reg_num + "$")
		self.assertRegex("-34.8", "^" + c.reg_num + "$")
		self.assertRegex("-94.889", "^" + c.reg_num + "$")
		self.assertRegex("954.189", "^" + c.reg_num + "$")
		self.assertRegex(".189", "^" + c.reg_num + "$")
		self.assertRegex("954.", "^" + c.reg_num + "$")
		self.assertRegex("-54.", "^" + c.reg_num + "$")
		self.assertRegex("-.3", "^" + c.reg_num + "$")

	def test_reg_num_does_not_match(self):
		self.assertNotRegex("text", "^" + c.reg_num + "$")
		self.assertNotRegex("1 . 2", "^" + c.reg_num + "$")
		self.assertNotRegex("1 .2", "^" + c.reg_num + "$")
		self.assertNotRegex("- 1.2", "^" + c.reg_num + "$")
		self.assertNotRegex("+ 1.2", "^" + c.reg_num + "$")

	def test_command_comp(self):
		'''
		Test that all the commands get matched by the regular
		expression.
		'''

		self.assertRegex("history", c.regular_expr["command_comp"])
		self.assertRegex("History", c.regular_expr["command_comp"])
		self.assertRegex("quit", c.regular_expr["command_comp"])
		self.assertRegex("Quit", c.regular_expr["command_comp"])
		self.assertRegex("exit", c.regular_expr["command_comp"])
		self.assertRegex("Exit", c.regular_expr["command_comp"])
		self.assertRegex("degree mode", c.regular_expr["command_comp"])
		self.assertRegex("Degree mode", c.regular_expr["command_comp"])
		self.assertRegex("degree Mode", c.regular_expr["command_comp"])
		self.assertRegex("Degree Mode", c.regular_expr["command_comp"])
		self.assertRegex("radian mode", c.regular_expr["command_comp"])
		self.assertRegex("Radian mode", c.regular_expr["command_comp"])
		self.assertRegex("radian Mode", c.regular_expr["command_comp"])
		self.assertRegex("Radian Mode", c.regular_expr["command_comp"])
		self.assertRegex("", c.regular_expr["command_comp"])

	def test_const_comp(self):
		'''
		Test that all the constants get matched by the regular
		expression.
		'''

		self.assertRegex("pi", c.regular_expr["const_comp"])
		self.assertRegex("π", c.regular_expr["const_comp"])
		self.assertRegex("e", c.regular_expr["const_comp"])
		self.assertRegex("ans", c.regular_expr["const_comp"])
		self.assertRegex("answer", c.regular_expr["const_comp"])
		self.assertRegex("e", c.regular_expr["const_comp"])
		self.assertRegex("tau", c.regular_expr["const_comp"])
		self.assertRegex("τ", c.regular_expr["const_comp"])
		self.assertRegex("phi", c.regular_expr["const_comp"])
		self.assertRegex("φ", c.regular_expr["const_comp"])

	def test_const_comp_not_match_e_in_a_word(self):
		'''
		Test that the constant e isn't matched if it is surrounded
		by other letters.
		'''

		self.assertNotRegex("ae", c.regular_expr["const_comp"])
		self.assertNotRegex("hare", c.regular_expr["const_comp"])
		self.assertNotRegex("wer", c.regular_expr["const_comp"])
		self.assertNotRegex("integral", c.regular_expr["const_comp"])
		self.assertNotRegex("derivative", c.regular_expr["const_comp"])
		self.assertNotRegex("solve", c.regular_expr["const_comp"])
		self.assertNotRegex("eval", c.regular_expr["const_comp"])

	def test_graph_comp(self):
		self.assertEqual(
			c.regular_expr["graph_comp"].search("graph x").group(1),
			"x")
		self.assertEqual(
			c.regular_expr["graph_comp"].search(
				"graph x and x^2").group(1),
			"x and x^2")
		self.assertEqual(
			c.regular_expr["graph_comp"].search(
				"graph x from 0 to 1").group(1),
			"x from 0 to 1")
		self.assertEqual(
			c.regular_expr["graph_comp"].search(
				"Graph 2*x").group(1),
			"2*x")

	def test_alg_comp(self):
		self.assertEqual(
			c.regular_expr["alg_comp"].search(
				"solve(x=2)").group(1),
			"(x=2)")
		self.assertEqual(
			c.regular_expr["alg_comp"].search(
				"Solve (x=3)").group(1),
			" (x=3)")

	def test_nested_alg_comp(self):
		self.assertEqual(
			c.regular_expr["alg_comp"].search(
				"solve(x=solve(x=4))").group(1),
			"(x=solve(x=4))")

	def test_eval_comp(self):
		m = c.regular_expr["eval_comp"].search("eval x at 1")
		self.assertEqual(m.group(1), "x")
		self.assertEqual(m.group(2), "1")

		m = c.regular_expr["eval_comp"].search("Eval 2*x at 3")
		self.assertEqual(m.group(1), "2*x")
		self.assertEqual(m.group(2), "3")

		m = c.regular_expr["eval_comp"].search("eval x^3 for .5")
		self.assertEqual(m.group(1), "x^3")
		self.assertEqual(m.group(2), ".5")

		m = c.regular_expr["eval_comp"].search("Eval 3*x for 11")
		self.assertEqual(m.group(1), "3*x")
		self.assertEqual(m.group(2), "11")

	@unittest.expectedFailure
	def test_nested_eval_comp(self):
		m = c.regular_expr["eval_comp"].search(
			"eval x at eval 2*x for 3")
		self.assertEqual(m.group(1), "x")
		self.assertEqual(m.group(2), "eval 2*x for 3")

	def test_der_comp(self):
		m = c.regular_expr["der_comp"].search(
			"derivative of x^2 at 2")
		self.assertEqual(m.group(1), "x^2")
		self.assertEqual(m.group(2), "2")
		
		m = c.regular_expr["der_comp"].search(
			"derivative of 3*x at 4")
		self.assertEqual(m.group(1), "3*x")
		self.assertEqual(m.group(2), "4")

		m = c.regular_expr["der_comp"].search(
			"derivative of 6*g at 1 with respect to g")
		self.assertEqual(m.group(1), "6*g")
		self.assertEqual(m.group(2), "1")
		self.assertEqual(m.group(3), " with respect to g")

	@unittest.expectedFailure
	def test_nested_der_comp(self):
		m = c.regular_expr["der_comp"].search(
			"derivative of 6*g at derivative of x^2 at 6 with "
			"respect to x with respect to g")
		self.assertEqual(m.group(1), "6*g")
		self.assertEqual(
			m.group(2),
			"derivative of x^2 at 6 with respect to x")
		self.assertEqual(m.group(3), " with respect to g")

	def test_int_comp(self):
		m = c.regular_expr["int_comp"].search(
			"integral 2*x dx 0 to 1")
		self.assertEqual(m.group(1), "2*x ")
		self.assertEqual(m.group(2), "x")
		self.assertEqual(m.group(3), "0")
		self.assertEqual(m.group(4), "1")
		
		m = c.regular_expr["int_comp"].search(
			"integral 2*x dx from 0 to 1")
		self.assertEqual(m.group(1), "2*x ")
		self.assertEqual(m.group(2), "x")
		self.assertEqual(m.group(3), "0")
		self.assertEqual(m.group(4), "1")
		
		m = c.regular_expr["int_comp"].search(
			"integrate 2*x dx 0 to 1")
		self.assertEqual(m.group(1), "2*x ")
		self.assertEqual(m.group(2), "x")
		self.assertEqual(m.group(3), "0")
		self.assertEqual(m.group(4), "1")

		m = c.regular_expr["int_comp"].search(
			"∫3*x dx 2 to 4")
		self.assertEqual(m.group(1), "3*x ")
		self.assertEqual(m.group(2), "x")
		self.assertEqual(m.group(3), "2")
		self.assertEqual(m.group(4), "4")

		m = c.regular_expr["int_comp"].search(
			"∫3*r dr 2 to 4")
		self.assertEqual(m.group(1), "3*r ")
		self.assertEqual(m.group(2), "r")
		self.assertEqual(m.group(3), "2")
		self.assertEqual(m.group(4), "4")

	@unittest.expectedFailure
	def test_int_comp_non_number_bounds(self):
		m = c.regular_expr["int_comp"].search(
			"integrate 2*x dx 0 to 1+1")
		self.assertEqual(m.group(1), "2*x ")
		self.assertEqual(m.group(2), "x")
		self.assertEqual(m.group(3), "0")
		self.assertEqual(m.group(4), "1+1")

	def test_comb_comp(self):
		self.assertRegex("C(5,4)", c.regular_expr["comb_comp"])
		self.assertRegex("P(5,4)", c.regular_expr["comb_comp"])

	def test_ave_comp(self):
		re = c.regular_expr["ave_comp"]

		self.assertRegex("Mean(4,3)", re)
		self.assertRegex("Median(4,3)", re)
		self.assertRegex("Mode(4,3)", re)
		self.assertRegex("mean(4,3)", re)
		self.assertRegex("median(4,3)", re)
		self.assertRegex("mode(4,3)", re)
		self.assertRegex("Max(4,3)", re)
		self.assertRegex("Min(4,3)", re)
		self.assertRegex("max(4,3)", re)
		self.assertRegex("min(4,3)", re)
		self.assertRegex("Ave(4,3)", re)
		self.assertRegex("ave(4,3)", re)
		self.assertRegex("average(4,3)", re)
		self.assertRegex("Stdev(4,3)", re)
		self.assertRegex("stdev(4,3)", re)

		m = c.regular_expr["ave_comp"].search("average(6,4)")
		self.assertEqual(m.group(1), "average")
		self.assertEqual(m.group(2), "(6,4)")

	def test_trig_comp(self):
		'''
		Test that all the single argument function get matched.
		'''

		for i in c.one_arg_funcs:
			with self.subTest(i = i):
				self.assertRegex(
					"%s(pi)" % i,
					c.regular_expr["trig_comp"])

	def test_gamma_comp(self):
		'''
		Test the gamma function regular expression.
		'''

		self.assertRegex("gamma(3)", c.regular_expr["gamma_comp"])
		self.assertRegex("Gamma(4)", c.regular_expr["gamma_comp"])
		self.assertRegex("Γ(3)", c.regular_expr["gamma_comp"])

	def test_log_comp(self):
		'''
		Test the logarithm regular expression.
		'''

		self.assertRegex("Log(4,2)", c.regular_expr["log_comp"])
		self.assertRegex("log(4,2)", c.regular_expr["log_comp"])
		self.assertRegex("Ln(2)", c.regular_expr["log_comp"])
		self.assertRegex("ln(4)", c.regular_expr["log_comp"])

	def test_mod2_comp(self):
		'''
		Test the mod function regular expression.
		'''

		self.assertRegex("mod(5,2)", c.regular_expr["mod2_comp"])
		self.assertRegex("Mod(5,2)", c.regular_expr["mod2_comp"])
		self.assertRegex("Mod(5,mod(8,6))", c.regular_expr["mod2_comp"])

	def test_abs_comp(self):
		'''
		Test the absolute value regular expression.
		'''

		self.assertRegex("|-1|", c.regular_expr["abs_comp"])
		self.assertRegex("|-1", c.regular_expr["abs_comp"])

	def test_paren_comp(self):
		'''
		Test the parentheses regular expression.
		'''

		self.assertRegex("(5+6)", c.regular_expr["paren_comp"])
		self.assertEqual(
			c.regular_expr["paren_comp"].search("(1+(2-1))").group(1),
			"2-1")

	def test_comma_comp(self):
		'''
		Test comma_comp.
		'''

		self.assertRegex("5,211", c.regular_expr["comma_comp"])
		self.assertNotRegex("5, 6", c.regular_expr["comma_comp"])

	def test_choose_comp(self):
		'''
		Test choose notation regular expression.
		'''

		self.assertRegex("5C3", c.regular_expr["choose_comp"])
		self.assertRegex("5P3", c.regular_expr["choose_comp"])
		self.assertRegex("5.4C3.1", c.regular_expr["choose_comp"])
		self.assertRegex("5.9P1.3", c.regular_expr["choose_comp"])

	def test_exp_comp(self):
		'''
		Test the exponents regular expression.
		'''

		self.assertRegex("2^3", c.regular_expr["exp_comp"])
		self.assertRegex("2**3", c.regular_expr["exp_comp"])
		self.assertRegex("2^3.1", c.regular_expr["exp_comp"])
		self.assertRegex("2**3.4", c.regular_expr["exp_comp"])
		self.assertRegex("2 ^ 3", c.regular_expr["exp_comp"])
		self.assertRegex("2 ** 3", c.regular_expr["exp_comp"])
		self.assertRegex("2 ^ 3.1", c.regular_expr["exp_comp"])
		self.assertRegex("2 ** 3.4", c.regular_expr["exp_comp"])

	def test_fact_comp(self):
		'''
		Test factorial regular expression.
		'''

		self.assertRegex("4!", c.regular_expr["fact_comp"])
		self.assertRegex(".7!", c.regular_expr["fact_comp"])
		self.assertRegex("4.2!", c.regular_expr["fact_comp"])
		self.assertRegex("-4.2!", c.regular_expr["fact_comp"])

	def test_mod_comp(self):
		'''
		Test the modulo operator regular expression.
		'''

		self.assertRegex("7 % 2", c.regular_expr["mod_comp"])
		
		m = c.regular_expr["mod_comp"].search("11%2")
		self.assertEqual(m.group(1), "11")
		self.assertEqual(m.group(2), "2")

	def test_per_comp(self):
		'''
		Test the percent regular expression.
		'''

		self.assertRegex("11%", c.regular_expr["per_comp"])

	def test_mult_comp(self):
		'''
		Test the multiplication and division regular expression.
		'''

		m = c.regular_expr["mult_comp"].search("11*2")
		self.assertEqual(m.group(1), "11")
		self.assertEqual(m.group(2), "*")
		self.assertEqual(m.group(3), "2")

		m = c.regular_expr["mult_comp"].search("11/2")
		self.assertEqual(m.group(1), "11")
		self.assertEqual(m.group(2), "/")
		self.assertEqual(m.group(3), "2")

		m = c.regular_expr["mult_comp"].search("10÷3")
		self.assertEqual(m.group(1), "10")
		self.assertEqual(m.group(2), "÷")
		self.assertEqual(m.group(3), "3")

	def test_add_comp(self):
		'''
		Test addition and subtraction regular expression.
		'''

		m = c.regular_expr["add_comp"].search("5-4")
		self.assertEqual(m.group(1), "5")
		self.assertEqual(m.group(2), "-")
		self.assertEqual(m.group(3), "4")

		m = c.regular_expr["add_comp"].search("1+3")
		self.assertEqual(m.group(1), "1")
		self.assertEqual(m.group(2), "+")
		self.assertEqual(m.group(3), "3")


class TestNonRepeatingLists(unittest.TestCase):
	'''
	Test that the NonRepeatingList class works as expected.
	'''

	def test_instantiate(self):
		nonlist = c.NonRepeatingList(1, 1, 2, 1)
		self.assertIsInstance(nonlist, c.NonRepeatingList)

	def test_str(self):
		nonlist = c.NonRepeatingList(1, 1, 2, 1)
		self.assertEqual(str(nonlist), "[1, 2, 1]")

	def test_indexing(self):
		nonlist = c.NonRepeatingList(1, 1, 2, 1)
		self.assertEqual(nonlist[0], 1)
		self.assertEqual(nonlist[-2], 2)
		self.assertEqual(nonlist[:2], [1, 2])

	def test_in(self):
		nonlist = c.NonRepeatingList(1, 1, 2, 1)
		self.assertIn(2, nonlist)

	def test_len(self):
		nonlist = c.NonRepeatingList(1, 1, 2, 1)
		self.assertEqual(len(nonlist), 3)

	def test_del(self):
		nonlist = c.NonRepeatingList(1, 1, 2, 1)
		del nonlist[1]
		self.assertEqual(len(nonlist), 1)

	def test_append(self):
		nonlist = c.NonRepeatingList(1)
		nonlist.append(4, 5, 5, 6)
		self.assertEqual(len(nonlist), 4)

	def test_clear(self):
		nonlist = c.NonRepeatingList(1, 1, 2, 1)
		nonlist.clear()
		self.assertEqual(nonlist, c.NonRepeatingList())


class SimplifyMonoFunction(unittest.TestCase):
	'''
	Test the simplify function.
	'''

	def test_adding(self):
		self.assertEqual(c.simplify("1+1"), "2")
		self.assertEqual(c.simplify("1+6"), "7")
		self.assertEqual(c.simplify("2.9+3"), "5.9")
		self.assertEqual(c.simplify("2.9+3.6 + 1 + 4.11"), "11.61")

	def test_subtracting(self):
		self.assertEqual(c.simplify("1-1"), "0")
		self.assertEqual(c.simplify("21-4"), "17")
		self.assertEqual(c.simplify("2-4"), "-2")
		self.assertEqual(c.simplify("2.5-3"), "-0.5")
		self.assertEqual(c.simplify("7-3-7"), "-3")

	def test_multipling(self):
		self.assertEqual(c.simplify("54.8 * 1 -1"), "53.8")
		self.assertEqual(c.simplify("54.8 * -2"), "-109.6")
		self.assertEqual(c.simplify("54.8 * -2"), "-109.6")
		self.assertEqual(c.simplify("2.5 * 2.5"), "6.25")
		self.assertEqual(c.simplify("2 - 2 * 4 * 9 + 5"), "-65")
		self.assertEqual(c.simplify("2(1+1)"), "4")
		self.assertEqual(c.simplify("6(5-2)"), "18")
		self.assertAlmostEqual(
			float(c.simplify("2pi")),
			math.tau)

	def test_dividing(self):
		self.assertEqual(c.simplify("1 / 1"), "1")
		self.assertEqual(c.simplify("1 / 4 + 2"), "2.25")
		self.assertEqual(c.simplify("27 ÷3"), "9")
		self.assertEqual(c.simplify("1 + 27 /-6"), "-3.5")
		self.assertEqual(c.simplify("27 ÷-6 / .25"), "-18")

	def test_percentages(self):
		self.assertEqual(c.simplify("15%"), "0.15")
		self.assertEqual(c.simplify("136%"), "1.36")
		self.assertEqual(c.simplify(".6%"), "0.006")
		self.assertEqual(c.simplify("4*5%"), "0.2")

	def test_modulo(self):
		self.assertEqual(c.simplify("8%7 + 1"), "2")
		self.assertEqual(c.simplify("9%7"), "2")
		self.assertEqual(c.simplify("9%4"), "1")
		self.assertEqual(c.simplify("11%3 - 4"), "-2")
		self.assertAlmostEqual(
			float(c.simplify("8.9%1")),
			0.9)
		self.assertAlmostEqual(
			float(c.simplify("3*8.2%5%2")),
			3.6)

	def test_factorial(self):
		self.assertEqual(c.simplify("5!"), "120")
		self.assertEqual(c.simplify("3!"), "6")
		self.assertEqual(c.simplify("3!!"), "720")
		self.assertAlmostEqual(
			float(c.simplify("4.5!")),
			52.3427777846)

	def test_exponents(self):
		self.assertEqual(c.simplify("5^2"), "25")
		self.assertEqual(c.simplify("3^3"), "27")
		self.assertEqual(c.simplify("9^1.5"), "27")
		self.assertEqual(c.simplify("4^.5"), "2")
		self.assertEqual(c.simplify("4**2.5"), "32")
		self.assertEqual(c.simplify("1**41"), "1")

	def test_choose_notation(self):
		self.assertEqual(c.simplify("5C2"), "10")
		self.assertEqual(c.simplify("5C3"), "10")
		self.assertEqual(c.simplify("5P2"), "20")
		self.assertEqual(c.simplify("(2+3)C2"), "10")
		self.assertAlmostEqual(
			float(c.simplify("4.3 P 1.2")),
			5.589326932606859)

	def test_comma_comp(self):
		'''
		Test that commas still work when called from inside simplify.
		'''

		self.assertEqual(c.simplify("1,433,123"), "1433123")
		self.assertEqual(c.simplify("1,433,123.6"), "1433123.6")

	def test_comma_raises_error(self):
		with self.assertRaises(c.CalculatorError):
			c.simplify("2,3445.6")

	def test_parrentheses(self):
		self.assertEqual(c.simplify("(1-(3+4(3-2)-2))*-4+2"), "18")
		self.assertEqual(c.simplify("(((5*3)))"), "15")

	def test_absolute_value(self):
		self.assertEqual(c.simplify("|3-5|"), "2")
		self.assertEqual(c.simplify("|1-|3-5||"), "1")
		self.assertEqual(c.simplify("-|3-5|"), "-2")
		self.assertEqual(c.simplify("||6||"), "6")
		self.assertEqual(c.simplify("|-2*|6-|4+3|*2|+2|"), "14")

	def test_mod2(self):
		self.assertEqual(c.simplify("Mod(5,3)"), "2")
		self.assertEqual(c.simplify("mod(8,6)"), "2")
		self.assertEqual(c.simplify("mod(9,4)"), "1")
		self.assertEqual(c.simplify("mod(16,2)"), "0")
		self.assertAlmostEqual(float(c.simplify("mod(11.3,1)")), 0.3)
		self.assertEqual(c.simplify("mod(9*4,4)"), "0")
		self.assertEqual(c.simplify("mod(Mod(11,8),2)"), "1")
		self.assertEqual(c.simplify("mod(16.5,12**0)"), "0.5")

	def test_logarithm(self):
		self.assertEqual(c.simplify("Log(9,3)-1"), "1")
		self.assertEqual(c.simplify("log(27,9)"), "1.5")
		self.assertEqual(c.simplify("Log(1000,10)"), "3")
		self.assertEqual(c.simplify("Log(64,2)"), "6")
		self.assertEqual(c.simplify("Log(e^3,e)"), "3")
		self.assertEqual(c.simplify("ln(e)"), "1")
		self.assertEqual(c.simplify("Ln(e)"), "1")

	def test_gamma(self):
		self.assertEqual(c.simplify("gamma(6)"), "120")
		self.assertEqual(c.simplify("Γ(6)"), "120")
		self.assertEqual(c.simplify("Gamma(5)*.5"), "12")
		self.assertEqual(c.simplify("2^(Gamma(5)/8)"), "8")

	def test_single_argument_functions(self):
		pass








if __name__ == '__main__':
	for i in range(5):
		print("")
	doctest.testmod(c)
	unittest.main()



# -*- coding: utf-8 -*-

'''
Tkinter tests for ReCalc.

testing_ReCalc_tkinter.py
'''

import unittest
import tkinter as tk


class NonRepeatingList(object):
	'''
	A mutable list that doesn't have two of the same element in a row.
	
	>>> repr(NonRepeatingList(3, 3, 4))
	'NonRepeatingList(*[3, 4])'
	'''

	def __init__(self, *args):
		if len(args) > 0:
			self.items = [args[0]]
			for i in args:
				if i != self.items[-1]:
					self.items.append(i)
		else:
			self.items = []

	def __getitem__(self, index):
		return(self.items[index])

	def __delitem__(self, index):
		del self.items[index]
		if index != 0:
			if self.items[index] == self.items[index - 1]:
				del self.items[index]

	def __contains__(self, item):
		return(item in self.items)

	def __len__(self):
		return(len(self.items))

	def __repr__(self):
		return("NonRepeatingList(*" + repr(self.items) + ")")

	def __str__(self):
		return(str(self.items))

	def __eq__(self, other):
		if isinstance(other, NonRepeatingList):
			if self.items == other.items:
				return(True)
		return(False)

	def append(self, *args):
		for item in args:
			if len(self.items) > 0:
				if self.items[-1] != item:
					self.items.append(item)
			else:
				self.items.append(item)

	def clear(self):
		self.items.clear()


import ReCalc as c


class HistorhLengthChanging(unittest.TestCase):
	
	def setUp(self):
		self.original_hist_len = c.hist_len		

	def test_change_hist_len(self):
		self.root = tk.Tk()
		self.entry = tk.Entry(self.root)
		self.root.update()
		self.entry.delete(0, "end")
		self.entry.insert(0, "20")
		self.root.update()
		c.change_hist_len(self.entry, self.root)
		self.assertEqual(c.hist_len, 20)
		self.assertEqual(c.calc_info["hist_len"], 20)
		
	def test_float_input(self):
		self.test_change_hist_len()
		self.root = tk.Tk()
		self.entry = tk.Entry(self.root)
		self.root.update()
		self.entry.delete(0, "end")
		self.entry.insert(0, 49.5)
		self.root.update()

		pre_hist_len = c.hist_len
		c.change_hist_len(self.entry, self.root)
		self.assertEqual(c.hist_len, pre_hist_len)
		self.assertEqual(c.calc_info["hist_len"], pre_hist_len)
		self.root.destroy()
		
	def test_other_string(self):
		self.test_change_hist_len()
		self.root = tk.Tk()
		self.entry = tk.Entry(self.root)
		self.root.update()
		self.entry.delete(0, "end")
		self.entry.insert(0, "hello")
		self.root.update()

		pre_hist_len = c.hist_len
		c.change_hist_len(self.entry, self.root)
		self.assertEqual(c.hist_len, pre_hist_len)
		self.assertEqual(c.calc_info["hist_len"], pre_hist_len)
		self.root.destroy()
		
	def test_string_int(self):
		self.root = tk.Tk()
		self.entry = tk.Entry(self.root)
		self.root.update()
		self.entry.delete(0, "end")
		self.entry.insert(0, "50")
		self.root.update()
		c.change_hist_len(self.entry, self.root)
		self.assertEqual(c.hist_len, 50)
		self.assertEqual(c.calc_info["hist_len"], 50)
		
	def tearDown(self):
		self.root = tk.Tk()
		self.entry = tk.Entry(self.root)
		self.root.update()
		self.entry.delete(0, "end")
		self.entry.insert(0, self.original_hist_len)
		self.root.update()
		c.change_hist_len(self.entry, self.root)

		
class ChangeDerApprox(unittest.TestCase):

	def setUp(self):
		self.original_der_approx = c.der_approx

	def test_change_der_approx_float(self):
		self.root = tk.Tk()
		self.entry = tk.Entry(self.root)
		self.root.update()
		self.entry.delete(0, "end")
		self.entry.insert(0, ".002")
		self.root.update()
		
		c.change_der_approx(self.entry, self.root)
		self.assertEqual(c.der_approx, .002)
		self.assertEqual(c.calc_info["der_approx"], .002)
		
	def test_der_approx_negative(self):
		self.root = tk.Tk()
		self.entry = tk.Entry(self.root)
		self.root.update()
		self.entry.delete(0, "end")
		self.entry.insert(0, "-.01")
		self.root.update()
		pre_der_approx = c.der_approx
		
		c.change_der_approx(self.entry, self.root)
		self.assertEqual(c.der_approx, pre_der_approx)
		self.assertEqual(c.calc_info["der_approx"], pre_der_approx)
		self.root.destroy()
		
	def test_der_approx_random_string(self):
		self.root = tk.Tk()
		self.entry = tk.Entry(self.root)
		self.root.update()
		self.entry.delete(0, "end")
		self.entry.insert(0, "word")
		self.root.update()
		pre_der_approx = c.der_approx

		c.change_der_approx(self.entry, self.root)
		self.assertEqual(c.der_approx, pre_der_approx)
		self.assertEqual(c.calc_info["der_approx"], pre_der_approx)
		self.root.destroy()


class ChangeGraphWinSet(unittest.TestCase):
	pass


class CartGraph(unittest.TestCase):
	
	def test_create_graph(self):
		g = c.CartGraph(
			xmin = -1, xmax = 1, ymin = 2, ymax = 3,
			wide = 40, high = 50)
		self.assertEqual(g.xmin, -1)
		self.assertEqual(g.xmax, 1)
		self.assertEqual(g.ymin, 2)
		self.assertEqual(g.ymax, 3)
		self.assertEqual(g.wide, 40)
		self.assertEqual(g.high, 50)
		g.root.destroy()
	
	def test_close_while_graphing(self):
		g = c.CartGraph(
			xmin = -1, xmax = 1, ymin = 2, ymax = 3,
			wide = 40, high = 50)
		g.draw("1")
		g.root.destroy()
		
	def test_graph_partial_undefined(self):
		g = c.CartGraph(
			xmin = -1, xmax = 1, ymin = 2, ymax = 3,
			wide = 100, high = 100)
		g.draw("x**.5")
		g.root.destroy()

		
class PolarGraph(unittest.TestCase):
	
	def test_create_graph(self):
		g = c.NumpyPolarGraph(
			xmin = -1, xmax = 1, ymin = 2, ymax = 3,
			theta_min = -2, theta_max = 6, wide = 40, high = 50)
		self.assertEqual(g.xmin, -1)
		self.assertEqual(g.xmax, 1)
		self.assertEqual(g.ymin, 2)
		self.assertEqual(g.ymax, 3)
		self.assertEqual(g.theta_min, -2)
		self.assertEqual(g.theta_max, 6)
		self.assertEqual(g.wide, 40)
		self.assertEqual(g.high, 50)
		g.root.destroy()
	
	def test_close_while_graphing(self):
		g = c.NumpyPolarGraph(
			xmin = -1, xmax = 1, ymin = 2, ymax = 3,
			theta_min = -.5, theta_max = .5, wide = 40, high = 50)
		g.draw("1")
		g.root.destroy()
		
	def test_graph_partial_undefined(self):
		g = c.NumpyPolarGraph(
			xmin = -1, xmax = 1, ymin = 2, ymax = 3,
			theta_min = -.5, theta_max = .5, wide = 100, high = 100)
		g.draw("x**.5")
		g.root.destroy()

		
class GraphFunction(unittest.TestCase):
	
	def setUp(self):
		self.mode = c.polar_mode
		c.switch_polar_mode(False)
		
	def test_basic_graph(self):
		g = c.graph_function("x")
		g.root.destroy()
		
	def test_graph_from(self):
		g = c.graph_function("x from 10 to 11")
		g.root.destroy()
		
	def test_multiple_graphs(self):
		g = c.graph_function("x and 2*x")
		g.root.destroy()
		
	def test_multiple_graphs_and_from(self):
		g = c.graph_function("x and 2*x from 10 to 11")
		g.root.destroy()

	def tearDown(self):
		c.switch_polar_mode(self.mode)


class GraphFunctionPolar(unittest.TestCase):
	
	def setUp(self):
		self.mode = c.polar_mode
		c.switch_polar_mode(True)
		
	def test_basic_graph(self):
		g = c.graph_function("x")
		g.root.destroy()
		
	def test_graph_from(self):
		g = c.graph_function("x from 10 to 11")
		g.root.destroy()
		
	def test_multiple_graphs(self):
		g = c.graph_function("x and 2*x")
		g.root.destroy()
		
	def test_multiple_graphs_and_from(self):
		g = c.graph_function("x and 2*x from 10 to 11")
		g.root.destroy()

	def tearDown(self):
		c.switch_polar_mode(self.mode)
































if __name__ == '__main__':
    unittest.main()



#   setup.py

import os
import pickle
from setuptools import setup, find_packages

setup(
    name = "ReCalc",
    version = "0.1.6",
    packages = find_packages(),
	scripts = [
		"ReCalc.py",
		"testing_ReCalc.py",
		"testing_ReCalc_tkinter.py"
	],
	
	classifiers = [
		"Programming Language :: Python :: 3.6",
		"Environment :: X11 Applications :: Gnome",
		"License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
		"Natural Language :: English",
		"Operating System :: Microsoft :: Windows :: Windows 10",
		"Operating System :: POSIX :: Linux",
		"Programming Language :: Python :: 3 :: Only",
		"Programming Language :: Python :: Implementation :: CPython",
		"Topic :: Scientific/Engineering :: Mathematics"
	],
	install_requires = [
		"sympy>=1.1.1",
		"pillow>=5.0.0"
	],
	
	package_data = {
	"": ["*.ico", "*.txt", "*.py"]
	},
	zip_safe = False,
	
	url = "https://github.com/RedKnite5/ReCalc.git",
	
	author = "Max Friedman",
	author_email = "mr.awesome10000@gmail.com",
	# desciption = "This is a basic graphing calculator.",
	# licencse = "GPLv3",
	keywords = ["Calculator"]
)

class NonRepeatingList(object):
	'''
	A mutable list that doesn't have two of the same element in a row.
	
	>>> repr(NonRepeatingList(3, 3, 4))
	'NonRepeatingList(*[3, 4])'
	'''

	def __init__(self, *args):
		if len(args) > 0:
			self.items = [args[0]]
			for i in args:
				if i != self.items[-1]:
					self.items.append(i)
		else:
			self.items = []

	def __getitem__(self, index):
		return(self.items[index])

	def __delitem__(self, index):
		del self.items[index]
		if index != 0:
			if self.items[index] == self.items[index - 1]:
				del self.items[index]

	def __contains__(self, item):
		return(item in self.items)

	def __len__(self):
		return(len(self.items))

	def __repr__(self):
		return("NonRepeatingList(*" + repr(self.items) + ")")

	def __str__(self):
		return(str(self.items))

	def __eq__(self, other):
		if isinstance(other, NonRepeatingList):
			if self.items == other.items:
				return(True)
		return(False)

	def append(self, *args):
		for item in args:
			if len(self.items) > 0:
				if self.items[-1] != item:
					self.items.append(item)
			else:
				self.items.append(item)

	def clear(self):
		self.items.clear()


with open("log_file.log", "w+"):
	pass
with open("ReCalc_info.txt", "wb+") as file:
	calc_info = {}
	calc_info["history"] = NonRepeatingList()
	calc_info["degree_mode"] = 0
	calc_info["polar_mode"] = False
	calc_info["der_approx"] = .0001
	calc_info["hist_len"] = 100
	calc_info["window_bounds"] = {
		"x min": -5,
		"x max": 5,
		"y min": -5,
		"y max": 5,
		"theta min": 0,
		"theta max": 10,
	}

	pickle.dump(calc_info, file)





# -*- coding: utf-8 -*-

'''
Author: Max Friedman
License: GPLv3
Title: ReCalc

This is a graphing calculator written in python 3.6.
It can do:
addition, subtraction, multiplication, division, exponents,
logarithms, trig functions, inverse trig functions, hyperbolic
functions, inverse hyperbolic functions, the ceiling function,
the floor function, the gauss error function, modulus, absolute value,
factorials, the gamma function, combinations, permutations, max, min,
mean, median, mode, sample standard deviation, definite integrals,
derivatives at a point, evaluating a function at a point, graphing
functions in Cartesian and polar, and solving equations.

It has a GUI made with tkinter that it will default to using if tkinter
is installed. It defines graphing and polar graphing classes. It
remembers every input it is given.

'''


#  Windows:
#  C:\Users\Max\Documents\Python\ReCalc\ReCalc.py
#  Bash:
#  C:/Users/Max/Documents/Python/ReCalc/ReCalc.py

import math
import statistics as stats
import sys
import os
import logging
import atexit

import numpy as np
from PIL import Image
from warnings import warn, simplefilter
from pickle import load, dump
from re import compile, sub
from decimal import Context

logging.basicConfig(
	format = '%(levelname)s:%(message)s',
	filename = "log_file.log",
	level = logging.INFO)
logging.info("Program is starting.")

# import sympy if installed
try:
	from sympy import symbols, integrate, sympify
	from sympy.solvers import solve
except ModuleNotFoundError:
	logging.warning("Sympy could not be imported.")
	simplefilter('default', ImportWarning)
	warn(
		("Sympy can not be imported. Solving equations and definite "
		"integrals will not be available."),
		category = ImportWarning)

# import tkinter if installed
try:
	import tkinter as tk
	from tkinter import filedialog
	from _tkinter import TclError
	from PIL import ImageTk
except ModuleNotFoundError:
	logging.warning("Tkinter could not be imported.")
	simplefilter('default', ImportWarning)
	warn(
		"Tkinter can not be imported. Using command line interface",
		category = ImportWarning)


'''  Completed
2) Make average functions deal with non-number arguments
3) Stdev
4) deal with commas
5) tkinter interface
8) min max functions
9) make tkinter windows close
12) two graphs at once
14) eval non-number arguments
15) multi argument functions cant have commas in the second argument
16) absolute value with pipes
17) use division sign
18) set y bounds on graphs
19) pickle degree mode
20) graph closes only when user dictates
28) improve tkinter interface
29) cut off trailing zeros
30) two expressions adjacent means multiplication
31) polar graphs
33) show request
37) other weird trig functions
40) doc strings for all of the tests that need them
43) save graphs
44) on fedora only one enter key is bound in the GUI
46) log errors
47) don't stop the program when there is an error
'''

'''  To Do
1) Deal with floating point errors
6) complex numbers
7) higher order derivatives
10) graph non-functions
11) improper integrals
13) integrals can have non-number bounds
14) derivatives non-number arguments
21) matrices
22) unit conversions
23) indefinite integrals
24) derivatives of functions
25) summation notation
26) big pi notation
27) series
32) 3d graphs
34) make icon of tkinter window when run on Fedora
35) make compatible with other operating systems
36) fix subtraction problem
38) setup and wheel files
39) make tests for all parts of the program
41) user defined functions
42) user defined variables
45) parametric functions
48) error when you close a polar graphing window early
49) don't let the user pass ln(x) multiple arguments
50)
'''


# os handling
if os.name == "nt":
	user_path = os.environ["USERPROFILE"]
elif os.name == "posix":
	user_path = os.environ["HOME"]

logging.info("operating system: %s", os.name)

up_hist = 0
ctx = Context()

# changeable variables
use_gui = True
graph_w = 400
graph_h = 400
graph_colors = ("black", "red", "blue", "green", "orange", "purple")
ctx.prec = 17

color_dict = {
	"black": (0, 0, 0),
	"red": (255, 0, 0),
	"green": (0, 128, 0),
	"blue": (0, 0, 255),
	"orange": (255, 165, 0),
	"purple": (128, 0, 128),
	"magenta": (255, 0, 255),
}

key_binds = {
	"nt": {13: "enter", 38: "up", 40: "down"},
	"posix": {104: "enter", 36: "enter", 111: "up", 116: "down"},
}

g_bound_names = (
	"x min", "x max",
	"y min", "y max",
	"theta min", "theta max"
)

one_arg_funcs = {
	#   name                 function                angles
	"hacovercosin": (lambda x: .5 + math.sin(x) / 2, "in"),
	"hacoversin": (lambda x: .5 - math.sin(x) / 2, "in"),
	"havercosin": (lambda x: .5 + math.cos(x) / 2, "in"),
	"haversin": (lambda x: .5 - math.cos(x) / 2, "in"),
	"covercosin": (lambda x: 1 + math.sin(x), "in"),
	"coversin": (lambda x: 1 - math.sin(x), "in"),
	"vercosin": (lambda x: 1 + math.cos(x), "in"),
	"versin": (lambda x: 1 - math.cos(x), "in"),
	"exsec": (lambda x: 1 / math.cos(x) - 1, "in"),
	"excsc": (lambda x: 1 / math.sin(x) - 1, "in"),

	"asinh": (math.asinh, "out"),
	"acosh": (math.acosh, "out"),
	"atanh": (math.atanh, "out"),
	"asech": (lambda x: math.acosh(1 / x), "out"),
	"acsch": (lambda x: math.asinh(1 / x), "out"),
	"acoth": (lambda x: math.atanh(1 / x), "out"),

	"arcsinh": (math.asinh, "out"),
	"arccosh": (math.acosh, "out"),
	"arctanh": (math.atanh, "out"),
	"arcsech": (lambda x: math.acosh(1 / x), "out"),
	"arccsch": (lambda x: math.asinh(1 / x), "out"),
	"arccoth": (lambda x: math.atanh(1 / x), "out"),

	"sinh": (math.sinh, "in"),
	"cosh": (math.cosh, "in"),
	"tanh": (math.tanh, "in"),
	"sech": (lambda x: 1 / math.cosh(x), "in"),
	"csch": (lambda x: 1 / math.sinh(x), "in"),
	"coth": (lambda x: 1 / math.tanh(x), "in"),

	"asin": (math.asin, "out"),
	"acos": (math.acos, "out"),
	"atan": (math.atan, "out"),
	"asec": (lambda x: math.acos(1 / x), "out"),
	"acsc": (lambda x: math.asin(1 / x), "out"),
	"acot": (lambda x: math.atan(1 / x), "out"),

	"arcsin": (math.asin, "out"),
	"arccos": (math.acos, "out"),
	"arctan": (math.atan, "out"),
	"arcsec": (lambda x: math.acos(1 / x), "out"),
	"arccsc": (lambda x: math.asin(1 / x), "out"),
	"arccot": (lambda x: math.atan(1 / x), "out"),

	"sin": (math.sin, "in"),
	"cos": (math.cos, "in"),
	"tan": (math.tan, "in"),
	"sec": (lambda x: 1 / math.cos(x), "in"),
	"csc": (lambda x: 1 / math.sin(x), "in"),
	"cot": (lambda x: 1 / math.tan(x), "in"),

	"abs": (math.fabs, ""),
	"ceil": (math.ceil, ""),
	"floor": (math.floor, ""),
	"erf": (math.erf, ""),
	"erfc": (math.erfc, "")
}

# regex for a number
reg_num = "(-?[0-9]+\.?[0-9]*|-?[0-9]*\.?[0-9]+)"

# regular expressions
regular_expr = dict(

	# regex for commands
	# ^$ is for an empty string
	command_comp = compile(
		"([Hh]istory)|([Qq]uit|[Ee]xit|^$)|"
		"([Dd]egree [Mm]ode)|([Rr]adian [Mm]ode)"),

	implicet_mult_pre_const_comp = compile(
		"([.0-9])(?=e|pi|π|tau|τ|phi|φ)"),

	# regex for constants
	const_comp = compile(
		"(pi|π|(?<![a-z0-9])e(?![a-z0-9])|"
		"ans(?:wer)?|tau|τ|phi|φ)"),

	implicet_mult_comp = compile(
		"([.0-9])(?=[(x])"),

	# regex for graphing
	graph_comp = compile("[Gg]raph (.+)"),

	# regex for equation solving
	alg_comp = compile("[Ss]olve(.+)"),

	# regex for evaluating functions
	eval_comp = compile("[Ee]val(?:uate)? (.+) (?:for|at) (.+)"),

	# regex for derivatives (must be at a number)
	der_comp = compile(
		"[Dd]erivative of (.+) at (.+?)"
		"( with respect to [a-z])?"),

	# regex for integrals (bounds must be numbers)
	int_comp = compile(
		"(?:[Ii]ntegra(?:te ?|l ?)|∫)(.+) ?d([a-z])"
		" (?:from )?" + reg_num + " to " + reg_num),

	# regex for combinations and permutations
	# parentheses is to differentiate it from choose notation
	comb_comp = compile("(C|P)(\(.+)"),

	# regex for statistics functions
	ave_comp = compile(
		"([Aa]verage|[Aa]ve|[Mm]ean|[Mm]edian|[Mm]ode|"
		"[Mm]ax|[Mm]in|[Ss]tdev)(.+)"),

	# regex for one argument functions
	# the order does matter because if trig functions come
	# before hyperbolic functions the "h" is interpreted as
	# part of the argument for the function
	trig_comp = compile(
		"("
		"hacovercosin|hacoversin|havercosin|haversin|"
		"covercosin|coversin|vercosin|versin|"
		"exsec|excsc|"
		"sinh|cosh|tanh|asinh|acosh|atanh|"
		"arcsinh|arccosh|arctanh|"
		"sech|csch|coth|asech|acsch|acoth|"
		"arcsech|arccsch|arccoth|"
		"sin|cos|tan|sec|csc|cot|"
		"asin|arcsin|acos|arccos|atan|arctan|"
		"asec|acsc|acot|arcsec|arccsc|arccot|"
		"abs|ceil|floor|erf"
		")(.+)"),

	# regex for gamma function
	gamma_comp = compile("(?:[Gg]amma|Γ)(.+)"),

	# regex for logarithms
	log_comp = compile("[Ll]og(.+)|[Ll]n(.+)"),

	# regex for modulus
	mod2_comp = compile("[Mm]od(.+)"),

	# regex for detecting absolute value
	abs_comp = compile("(.*\|.*)"),

	# here is where the order of operations starts to matter
	# it goes: parentheses, choose notation(nCm), exponents, factorial,
	# modulus, multiplication, addition

	# regex for parentheses
	# [^()] makes it only find the inner most parentheses
	paren_comp = compile("\(([^()]+)\)"),

	# ignores commas in the middle of numbers
	# could be problematic if two floats ever
	# end up next to each other
	comma_comp = compile(reg_num + "," + reg_num),

	# regex for choose notation (not recursive)
	# in the form of "nCm" or "nPm"
	choose_comp = compile(reg_num + " ?(C|P) ?" + reg_num),

	# regex for exponents (not recursive)
	exp_comp = compile(reg_num + " ?(\*\*|\^) ?" + reg_num),

	# regex for factorials (not recursive)
	fact_comp = compile(reg_num + "\!"),

	# regex in the form x % y (not recursive)
	mod_comp = compile(reg_num + " ?% ?" + reg_num),

	# regex for percentages (should probably be done without regex)
	per_comp = compile("%"),

	# regex for multiplication (not recursive)
	mult_comp = compile(reg_num + " ?([*/÷]) ?" + reg_num),

	# regex for addition (not recursive)
	add_comp = compile(reg_num + " ?([+-]) ?" + reg_num),
)


class CalculatorError(Exception):
	pass


class NonRepeatingList(object):
	'''
	A mutable list that doesn't have two of the same element in a row.
	
	>>> repr(NonRepeatingList(3, 3, 4))
	'NonRepeatingList(*[3, 4])'
	'''

	def __init__(self, *args):
		if len(args) > 0:
			self.items = [args[0]]
			for i in args:
				if i != self.items[-1]:
					self.items.append(i)
		else:
			self.items = []

	def __getitem__(self, index):
		return(self.items[index])

	def __delitem__(self, index):
		del self.items[index]
		if index != 0:
			if self.items[index] == self.items[index - 1]:
				del self.items[index]

	def __contains__(self, item):
		return(item in self.items)

	def __len__(self):
		return(len(self.items))

	def __repr__(self):
		return("NonRepeatingList(*" + repr(self.items) + ")")

	def __str__(self):
		return(str(self.items))

	def __eq__(self, other):
		if isinstance(other, NonRepeatingList):
			if self.items == other.items:
				return(True)
		return(False)

	def append(self, *args):
		for item in args:
			if len(self.items) > 0:
				if self.items[-1] != item:
					self.items.append(item)
			else:
				self.items.append(item)

	def clear(self):
		self.items.clear()


class Graph(object):
	'''
	Base class for all graphs.
	'''

	def __init__(
		self,
		xmin = -5, xmax = 5, ymin = -5, ymax = 5,
		wide = 400, high = 400):
		'''
		Initialize the graphing window.
		'''

		self.root = tk.Toplevel()

		self.root.title("ReCalc")

		# sets bounds
		self.xmin = xmin
		self.xmax = xmax
		self.ymin = ymin
		self.ymax = ymax

		self.xrang = self.xmax - self.xmin
		self.yrang = self.ymax - self.ymin

		# dimensions of the window
		self.wide = wide
		self.high = high

		# create the canvas
		self.screen = tk.Canvas(
			self.root,
			width = self.wide, height = self.high)
		self.screen.pack()
		
		self.options = tk.Menu(self.root)
		self.file_options = tk.Menu(self.options, tearoff = 0)
		
		self.file_options.add_command(
			label = "Save",
			command = self.save_image)
		self.file_options.add_command(
			label = "Exit",
			command = self.root.destroy)
		
		self.options.add_cascade(
			label = "File",
			menu = self.file_options)
		
		self.root.config(menu = self.options)
		

	def axes(self):
		'''
		Draw the axis.
		'''

		# adjusted y coordinate of x-axis
		b = self.high + (self.ymin * self.high / self.yrang)

		# adjusted x coordinate of y-axis
		a = -1 * self.xmin * self.wide / self.xrang

		try:
			# draw x-axis
			self.screen.create_line(0, b, self.wide, b, fill = "gray")

			# draw y-axis
			self.screen.create_line(a, self.high, a, 0, fill = "gray")

			self.root.update()
		except TclError as e:
			pass

	def save_image(self):
		'''
		Save the image to a file.
		'''

		fout = filedialog.asksaveasfile()
		self.image.save(fout)
		print("Saved")


class NumpyGraph(Graph):
	'''
	Cartesian Graphing window class using numpy and PIL.
	'''

	def __init__(
		self,
		xmin = -5, xmax = 5, ymin = -5, ymax = 5,
		wide = 400, high = 400):
		'''
		Initialize the graphing window.
		'''

		super().__init__(
			xmin = -5, xmax = 5, ymin = -5, ymax = 5,
			wide = 400, high = 400)

		self.data = np.zeros(
			(self.high, self.wide, 3),
			dtype = np.uint8)
		self.data.fill(255)

		global pic

		# create the image
		pic = ImageTk.PhotoImage(
			Image.fromarray(self.data, "RGB"))
		self.screen.create_image(
			self.wide / 2, self.high / 2, image = pic)

		# draws the axes
		self.axes()

	# draw the axes
	def axes(self):
		'''
		Draw the axis.
		'''

		global pic

		# adjusted y coordinate of x-axis
		b = self.high + (self.ymin * self.high / self.yrang)

		# adjusted x coordinate of y-axis
		a = -1 * self.xmin * self.wide / self.xrang

		try:
			self.data[int(round(b, 0)), :, :] = 0
		except IndexError:
			pass
		try:
			self.data[:, int(round(a, 0)), :] = 0
		except IndexError:
			pass

		pic = ImageTk.PhotoImage(Image.fromarray(self.data, "RGB"))
		self.screen.create_image(
			self.wide / 2, self.high / 2, image = pic)

		try:
			self.root.update()
		except TclError:
			pass

	def draw(self, func, color = "black"):
		'''
		Draw a Cartesian function.
		'''

		global pic

		pixel_color = color_dict[color]

		for i in range(self.data.shape[1]):

			x = i * self.xrang / self.wide + self.xmin
			y = float(evaluate(func, str(x)))
			a = int(round((x - self.xmin) * self.wide / self.xrang, 0))
			b = int(round(
				self.high - (y - self.ymin) * self.high / self.yrang,
				0))

			if 0 < b and b < self.high:
				self.data[b, i] = pixel_color

			self.image = Image.fromarray(self.data, "RGB")
			pic = ImageTk.PhotoImage(self.image)
			self.screen.create_image(
				self.wide / 2, self.high / 2, image = pic)

			self.root.update()


class NumpyPolarGraph(NumpyGraph):
	'''
	A polar graph using PIL and Numpy.
	'''

	def __init__(
		self,
		xmin = -5, xmax = 5, ymin = -5, ymax = 5,
		theta_min = 0, theta_max = 10,
		wide = 400, high = 400):
		
		super().__init__(
			xmin = -5, xmax = 5, ymin = -5, ymax = 5,
			wide = 400, high = 400)

		self.theta_min = theta_min
		self.theta_max = theta_max

		self.theta_rang = self.theta_max - self.theta_min

	def draw(self, func, color = "black"):
		'''
		Draw a polar function with numpy.
		'''

		global pic

		density = 1000
		theta = self.theta_min
		
		pixel_color = color_dict[color]

		while theta < self.theta_max:

			# move theta a little
			theta += self.theta_rang / density
			try:
				# eval the function at theta and set that to r
				r = float(evaluate(func, str(theta)))

				# find the slope at the point using find_derivative
				slope = float(find_derivative(func, str(theta)))

				x = r * math.cos(theta)
				y = r * math.sin(theta)

				# calculate how dense the points need to be
				# this function is somewhat arbitrary
				density = int((400 * math.fabs(slope)) + 500)

				# check if the graph goes off the screen
				if y > self.ymax or y < self.ymin or \
					x > self.xmax or x < self.xmin:
					denstiy = 2000

				# adjust coordinate for the
				# screen (this is the hard part)
				a = int(round(
					(x - self.xmin) * self.wide / self.xrang,
					0))
				b = int(round(
						self.high - (
							(y - self.ymin) * self.high / self.yrang),
						0))

				# draw the point
				if 0 < b and b < self.high and 0 < a and a < self.wide:
					self.data[b, a] = pixel_color
				self.image = Image.fromarray(self.data, "RGB")
				pic = ImageTk.PhotoImage(self.image)
				self.screen.create_image(
					self.wide / 2, self.high / 2, image = pic)

			except (ValueError, TclError) as e:
				pass

			# update the screen
			try:
				self.root.update()
			except TclError as e:
				theta = self.theta_max + 1


# multi session variables
try:
	calc_path = os.path.abspath(os.path.dirname(__file__))
	with open(
		os.path.join(calc_path, "ReCalc_info.txt"),
		"rb") as file:
		calc_info = load(file)
	history = calc_info["history"]
	ans = calc_info["ans"]
	# in degree mode 0 = off 2 = on
	degree_mode = calc_info["degree_mode"]
	polar_mode = calc_info["polar_mode"]
	der_approx = calc_info["der_approx"]
	hist_len = calc_info["hist_len"]
	win_bound = calc_info["window_bounds"]
except:
	raise CalculatorError("Loading settings failed.")

	'''
	history = NonRepeatingList()
	ans = 0
	degree_mode = 0
	polar_mode = False
	der_approx = .0001
	hist_len = 100
	'''


def log_end():
	logging.info("Program is ending.")

atexit.register(log_end)


def float_to_str(f):
	'''
	Convert the given float to a string,
	without resorting to scientific notation.

	>>> float_to_str(3.0030)
	'3.003'
	'''

	d1 = ctx.create_decimal(repr(float(f)))
	string = format(d1, 'f')
	if string[-2:] == ".0":
		return string[:-2]
	return string
	

def check_if_float(x):
	'''
	Test if a object can be made a float.

	>>> check_if_float("4.5")
	True

	>>> check_if_float("this")
	False
	'''

	try:
		float(x)
		return(True)
	except (ValueError, TypeError):
		return(False)


def save_info(**kwargs):
	'''
	Save settings to a file.
	'''

	if len(history) < 100:
		raise CalculatorError("The history was about to be deleted.")

	calc_info.update(kwargs)

	with open(os.path.join(
		calc_path,
		"ReCalc_info.txt"), "wb") as file:
		dump(calc_info, file)

	
def switch_degree_mode(mode):
	'''
	Switch between degree mode and radian mode.
	'''

	global degree_mode

	if mode == "degree":
		degree_mode = 2
	elif mode == "radian":
		degree_mode = 0
	elif mode in (0, 2):
		degree_mode = mode
	else:
		raise CalculatorError("Can not set degree_mode to '%s'" % mode)

	save_info(degree_mode = degree_mode)


def switch_polar_mode(mode):
	'''
	Switch between polar and Cartesian graphing.
	'''

	global polar_mode

	if mode == "polar":
		polar_mode = True
	elif mode == "Cartesian":
		polar_mode = False
	elif mode in (True, False):
		polar_mode = mode
	else:
		raise CalculatorError("Can not set polar_mode to '%s'" % mode)

	save_info(polar_mode = polar_mode)


def change_hist_len(entry_box, root):
	'''
	Change the length of the history print back.
	'''

	global hist_len

	# get user input
	input = entry_box.get()

	# if the input is a digit set that to be the
	# history print back length save and close the window
	if input.isdigit() and int(input) > 0:
		hist_len = int(input)
		save_info(hist_len = hist_len)

		root.destroy()
	else:
		pass


def change_hist_len_win():
	'''
	Create a popup to change the length of the
	history print back.
	'''

	root = tk.Toplevel()

	# create window text
	disp = tk.Message(
		root,
		text = "Current History print back length: " + str(hist_len))
	disp.grid(row = 0, column = 0)

	# create the input box
	entry_box = tk.Entry(root)
	entry_box.grid(row = 1, column = 0)

	# bind enter to setting the input to be the history length
	root.bind(
		"<Return>",
		lambda event: change_hist_len(entry_box, root))

	root.mainloop()


def change_der_approx(entry_box, root):
	'''
	Change the length of the history print back.
	'''

	global der_approx

	# get user input
	input = entry_box.get()

	# if the input is a positive float set that to be the
	# der_approx value save and close the window
	if check_if_float(input) and "-" not in input:
		der_approx = float(input)
		save_info(der_approx = der_approx)

		root.destroy()
	else:
		pass


def change_der_approx_win():
	'''
	Create a popup to change the length of the
	history print back.
	'''

	root = tk.Toplevel()

	# create window text
	disp = tk.Message(
		root,
		text = "Current der approx: " + str(der_approx))
	disp.grid(row = 0, column = 0)

	# create the input box
	entry_box = tk.Entry(root)
	entry_box.grid(row = 1, column = 0)

	# bind enter to setting the input to be the history length
	root.bind(
		"<Return>",
		lambda event: change_der_approx(entry_box, root))

	root.mainloop()


def change_graph_win_set():
	'''
	Change the graphing window bounds.
	'''

	global win_bound, g_bound_entry, g_bound_string

	g_bound_input = {}
	for i in g_bound_names:
		g_bound_input[i] = g_bound_entry[i].get()

	for i in g_bound_names:
		if check_if_float(g_bound_input[i]):
			win_bound[i] = float(g_bound_input[i])

	save_info(win_bound = win_bound)

	for i in g_bound_names:
		g_bound_string[i].set("%s = %s" % (i, win_bound[i]))


def find_match(s):
	'''
	Split a string into two parts. The first part contains the
	string until the first time parentheses are completely matched. The
	second part contains the rest of the string.

	>>> find_match("(4, (5), 3) + 2")
	('(4, (5), 3)', ' + 2')
	'''

	x = 0
	if not s.startswith("("):
		raise CalculatorError(
			"Error '%s' is an invalid input, must start with '('" % s)

	for i in range(len(s)):

		# count the parentheses
		if s[i] == "(":
			x += 1
		elif s[i] == ")":
			x -= 1

		if x == 0:

			# left is all the excess characters after
			# the matched parentheses
			# an is the matched parentheses and everything in them
			an = s[:i+1]
			left = s[i+1:]

			break

		elif x < 0:
			raise CalculatorError(
				"'%s' is an invalid input. Parentheses do "
				"not match." % s)

	try:
		return(an, left)
	except UnboundLocalError:
		raise CalculatorError(
			"'%s' is an invalid input to find_match." % s)


def brackets(s):
	'''
	Return True if the parentheses match, False otherwise.

	>>> brackets("())")
	False

	>>> brackets("(()())")
	True

	>>> brackets("w(h)(a()t)")
	True
	'''

	x = 0
	for i in s:
		if i == "(":
			x += 1
		elif i == ")":
			x -= 1
		if x < 0:
			return(False)
	return(not x)


def separate(s):
	'''
	Split up arguments of a function with commas
	like mod(x, y) or log(x, y) based on where commas that aren't in
	parentheses.

	>>> separate("5, 6 , (4, 7)")
	('5', ' 6 ', ' (4, 7)')
	'''

	terms = s.split(",")

	new_terms = []
	middle = False
	term = ""
	for i in terms:
		if middle:
			term = term + "," + i
		else:
			term = i
		x = brackets(term)
		if x:
			new_terms.append(term)
			middle = False
		else:
			middle = True
	return(tuple(new_terms))


#####################
# List of Functions #
#####################

def constant(constant):
	'''
	Evaluate mathematical constants.

	>>> constant("pi")
	3.141592653589793

	>>> constant("phi")
	1.618033988749895
	'''

	constant_dict = {
		"pi": math.pi, "π": math.pi, "e": math.e,
		"ans": ans, "answer": ans, "tau": math.tau, "τ": math.tau,
		"phi": (1 + 5 ** 0.5) / 2, "φ": (1 + 5 ** 0.5) / 2}

	if constant in constant_dict:
		return(constant_dict[constant])
	else:
		raise CalculatorError(
			"'%s' is not a recognized constant." % constant)


def graph_function(func_arg):
	'''
	Graph the given function or functions.
	'''

	# check for bounds on graph
	graph_rang_comp = compile(
		"(.+(?=from))(from " + reg_num + " to " + reg_num + ")")

	# looks for x bounds on the graph
	range_m = graph_rang_comp.search(func_arg)
	if range_m is not None:
		func_arg = range_m

	# checks to see if tkinter is installed to graph things at all
	if "tkinter" in sys.modules:

		# finds multiple functions to graph
		if range_m is None:
			funcs_to_graph = func_arg.split(" and ")
		else:
			funcs_to_graph = func_arg.group(1).split(" and ")

		# sets bounds to bounds given
		if range_m is None:
			temp_graph_xmin = win_bound["x min"]
			temp_graph_xmax = win_bound["x max"]
		else:
			temp_graph_xmin = float(func_arg.group(3))
			temp_graph_xmax = float(func_arg.group(4))

		# creates graph object
		if polar_mode is False:
			made_graph = NumpyGraph(
				xmin = temp_graph_xmin, xmax = temp_graph_xmax,
				ymin = win_bound["y min"], ymax = win_bound["y max"],
				wide = graph_w, high = graph_h)
		else:
			made_graph = NumpyPolarGraph(
				xmin = temp_graph_xmin, xmax = temp_graph_xmax,
				ymin = win_bound["y min"], ymax = win_bound["y max"],
				theta_min = win_bound["theta min"],
				theta_max = win_bound["theta max"],
				wide = graph_w, high = graph_h)

		# works out how many times it needs to
		# loop the colors its using
		color_loops = math.ceil(
			len(funcs_to_graph) / len(graph_colors))

		# passes functions to be graphed and the color to do so with
		for f, c in zip(funcs_to_graph, graph_colors * color_loops):
			made_graph.draw(f, color = c)

		# necessary for some of the tests
		return(made_graph)

	# informs the user of reason for failure
	else:
		raise CalculatorError(
			"Could not graph. Tkinter is not installed")


def solve_equations(equation):
	'''
	Solve equations using sympy. If there is no equals
	sign it is assumed the expression equals zero. If there is more
	than one answer it will return a list of the answers.

	>>> solve_equations("x^2 = 4")
	[-2, 2]

	>>> solve_equations("x-7 = 13")
	20
	'''

	if "sympy" in sys.modules:
		# check for equals sign when solving equations
		eq_sides_comp = compile("(.+)=(.+)|(.+)")

		# checks for specified variable when solving equations
		alg_var_comp = compile("(.+) for ([a-z])")

		# find if there is a specified variable
		varm = alg_var_comp.search(equation)

		# find the variable its solving for. defaults to "x"
		if varm is None:
			x = symbols("x")
			eq = equation
		else:
			# used to be symbols(varm.group(2)[-1]) don't know why
			# may have been important
			x = symbols(varm.group(2))
			eq = varm.group(1)

		# if there is an equals sign solve for zero and use sympy
		# to solve it
		em = eq_sides_comp.search(eq)
		if em.group(3) is None:
			try:
				sym_zero = sympify(
					em.group(1) + "-(" + em.group(2) + ")")
				temp_result = solve(sym_zero, x)
			except (NotImplementedError, TypeError) as e:
				raise CalculatorError(str(e))

			# if there is only one result make it the result
			# otherwise return the list
			if len(temp_result) == 1:
				return(temp_result[0])
			else:
				return(temp_result)

		# if there isn't an equals sign use sympy to solve
		else:
			temp_result = solve(em.group(3), x)

			# if there is only one result make it the result
			# otherwise return the list
			if len(temp_result) == 1:
				return(temp_result[0])
			else:
				return(temp_result)
	else:
		raise CalculatorError(
			"Could not solve. Sympy not installed.")


def evaluate(expression, point, var = "x"):
	'''
	Evaluate the function by substituting var for the number you
	want to evaluate at.

	>>> evaluate("r^2", 5, var = "r")
	'25'

	>>> evaluate("3*b", "2", "b")
	'6'
	'''

	# substituting the point for x in the function and evaluating
	# that recursively
	return(simplify(sub(
		"(?<![a-z])" + var + "(?![a-z])",
		str(point), expression)))


def find_derivative(expression, point, var = "x"):
	'''
	Calculate the derivative by evaluating the slope
	between two points on either side of the point you are
	finding the derivative of.
	'''

	# find the point on either side of the desired point
	p = float(simplify(point))
	x_one = p + der_approx
	x_two = p - der_approx

	# find the change in y value between the two points
	delta_y = (float(evaluate(
		expression, str(x_one), var = var))
		- float(evaluate(expression, str(x_two), var = var)))

	# divide by the length of the interval to find the slope
	return(delta_y / (2 * der_approx))


def integrate_function(expression, var, a, b):
	'''
	Integrate with sympy.
	'''

	# Integrals must be in a form that sympy can integrate
	# The bounds must be numbers not expressions
	# The integral must have a "dx" or whatever variable you are using
	# using sympy to integrate
	if "sympy" in sys.modules:
		return(float_to_str(integrate(expression, (var, a, b))))
	raise CalculatorError(
		"Could not integrate sympy is not installed.")


def combinations_and_permutations(form, letter, n, m = None):
	'''
	Solve combinations and permutations.

	n is either n or all arguments depending of the for they're
	written in

	combinations and permutations written
	both as C(5, 2) and 5C2 evaluated as: 5 choose 2
	if written as mCn it will only take numbers not expressions
	unless parentheses are used. In order of operations nCm comes
	first.
	Combinations and permutations both used the gamma function
	in place of factorials and as a result will take
	non-integer arguments
	'''

	if form == "choose":  # if written as nCm

		if m is None:
			raise CalculatorError(
				"combinations_and_permutations requires a fourth "
				"argument when using choose notation")

		# turn the arguments into floats and give them more
		# descriptive names
		inner_n = float(n)
		inner_m = float(m)

		# find permutations
		temp_result = math.gamma(1 + inner_n) \
			/ math.gamma(1 + inner_n - inner_m)

		# if combinations also divide by m!
		if letter == "C":
			temp_result = temp_result / math.gamma(1 + inner_m)
		elif letter == "P":
			pass
		else:
			raise CalculatorError(
				"Second argument must be 'C' or 'P'"
				" for combinations or permutations not '%s'" % letter)

		return(str(temp_result))

	elif form == "func":  # if written as C(5, 2)
		if m:
			raise CalculatorError(
				"combinations_and_permutations can not take a fourth "
				"argument when using function notation")

		# find the arguments of the function and cut off
		# everything else
		# sin(C(5, 2)) ← the last parenthesis
		proto_inner = find_match(n)

		# remove outer parentheses
		x = proto_inner[0][1:-1]

		# separate the arguments
		comb_args = separate(x)
		if len(comb_args) != 2:
			raise CalculatorError(
				"combinations_and_permutations takes exactly "
				"two arguments %s were provided." % len(comb_args))

		# evaluate each of the arguments separately
		inner_n = float(simplify(comb_args[0]))
		inner_m = float(simplify(comb_args[1]))

		# find permutations
		temp_result = math.gamma(1 + inner_n) \
			/ math.gamma(1 + inner_n - inner_m)

		# if combinations also divide by m!
		if letter == "C":
			temp_result = temp_result / math.gamma(1 + inner_m)
		elif letter == "P":
			pass
		else:
			raise CalculatorError(
				"Second argument must be 'C' or 'P'"
				" for combinations or permutations not '%s'" % letter)

		# add on anything that was was cut off the end when finding
		# the arguments
		# sin(C(5, 2)) ← the last parenthesis
		return(str(temp_result) + proto_inner[1])
	else:
		raise CalculatorError(
			"First argument must be 'choose' or "
			"'func' not '%s'" % form)


def statistics_functions(function, args):
	'''
	Perform general statistics functions.

	This may in the future include any function that
	can have an arbitrarily large number of arguments.
	'''

	# find the arguments of the function and cut off
	# everything else
	# sin(mean(4, 2)) ← the last parenthesis
	proto_inner = find_match(args)

	# separate the arguments based on commas that are not
	# within more than one set of parentheses
	ave_args = separate(proto_inner[0][1:-1])

	try:
		# evaluate all the arguments
		list_ave = list(map((lambda x: float(simplify(x))), ave_args))
	except ValueError as e:
		raise CalculatorError(str(e))

	# perform the different functions
	if function.lower() in ("ave", "average", "mean"):
		result = stats.mean(list_ave)
	elif function.lower() == "median":
		result = stats.median(list_ave)
	elif function.lower() == "mode":
		result = stats.mode(list_ave)
	elif function.lower() == "max":
		result = max(list_ave)
	elif function.lower() == "min":
		result = min(list_ave)
	# this is the sample standard deviation
	elif function.lower() in ("stdev"):
		result = stats.stdev(list_ave)
	else:
		raise CalculatorError(
			"'%s' is not a function that is defined in"
			" statistics_functions" % function)

	# add on anything that was was cut off the end when finding
	# the arguments
	# sin(mean(4, 2)) ← the last parenthesis
	return(float_to_str(result) + proto_inner[1])


def single_argument(func, args):
	'''
	Evaluate trig functions and other unary operators.
	'''

	global degree_mode

	# find the arguments of the function and cut off
	# everything else
	# tan(sin(π)) ← the last parenthesis when
	# evaluating sin
	proto_inner = find_match(args)
	proto_inner = list(proto_inner)

	# looks for the degree symbol in the argument
	# if the program finds it degree mode is set to true
	# for the particular operation
	if "°" in proto_inner[0]:
		if degree_mode == 0:
			degree_mode = 1
		proto_inner[0] = sub("[°]", "", proto_inner[0])

	# evaluate the argument into a form that math.log
	# can accept
	pre_inner = simplify(proto_inner[0])
	try:
		inner = float(pre_inner)
	except ValueError as e:
		raise CalculatorError(str(e))
	

	# check if in degree mode and if its doing an
	# operation that takes an angle as an argument
	if degree_mode > 0 and "in" in one_arg_funcs[func][1]:
		inner = math.pi * inner / 180

	result = one_arg_funcs[func][0](inner)

	# checks if its in degree mode (not because of
	# degree symbols in the argument) and if so
	# converts the answer to degrees for functions that
	# output an angle
	if degree_mode == 2 and "out" in one_arg_funcs[func][1]:
		result = result * 180 / math.pi

	# resets the degree mode for the session
	if degree_mode == 1:
		degree_mode = 0

	# this is a fix for the output being in
	# scientific notation and the program mistaking the
	# e for the constant e
	try:
		result = float_to_str(result)
	except (ValueError, TypeError):
		pass

	# add back anything that was cut off when finding the
	# argument of the inner function
	# tan(sin(π)) ← the last parenthesis when
	# evaluating sin
	return(result + proto_inner[1])


def gamma(arg):
	'''
	Use the gamma function.

	>>> gamma("(4)")
	'6'

	>>> gamma("(4.6)")
	'13.381285870932441'
	'''

	# find the arguments of the function and cut off
	# everything else
	# sin(gamma(5)) ← the last parenthesis
	proto_inner = find_match(arg)

	# evaluating the argument
	pre_inner = simplify(proto_inner[0])
	try:
		inner = float(pre_inner)
	except ValueError as e:
		raise CalculatorError(str(e))

	result = math.gamma(inner)

	# add back anything that was cut off when finding the
	# argument of the inner function
	# sin(gamma(5)) ← the last parenthesis
	return(float_to_str(result) + proto_inner[1])


def factorial(arg):
	'''
	Evaluate factorials.

	Interprets x! mathematically as gamma(x + 1)
	if written with an "!" will only take numbers as an argument.
	In order of operations factorials come after exponents,
	but before modulus

	>>> factorial("5")
	120.0
	'''

	return(math.gamma(float(arg) + 1))


def logarithm(log_arg):
	'''
	Solve logarithms.

	Can be written log(a, b) or ln(x). Ln only takes one argument
	and is evaluated separately from log.
	'''

	# find the arguments of the function and cut off
	# everything else
	# sin(log(4, 2)) ← the last parenthesis
	proto_inner = find_match(log_arg)

	# separate the arguments based on commas that are not
	# within more than one set of parentheses
	log_args = separate(proto_inner[0][1:-1])

	if len(log_args) == 2:  # if written as log(a, b)

		# evaluate the arguments individually into a form
		# that math.log can accept
		argument = float(simplify(log_args[0]))
		base = float(simplify(log_args[1]))

		# perform the logarithm
		if base == 2:
			result = math.log2(argument)
		elif base == 10:
			result = math.log10(argument)
		elif base == math.e:
			result = math.log(argument)
		else:
			result = math.log(argument, base)

		return(str(result) + proto_inner[1])

	elif len(log_args) == 1:  # if written as ln(x)

		result = math.log(float(simplify(proto_inner[0])))

	else:
		raise CalculatorError(
			"There must be 1 or 2 arguments in the "
			"input string")

	# add on anything that was was cut off the end when finding
	# the arguments
	# sin(log(4, 2)) ← the last parenthesis
	return(str(result) + proto_inner[1])


def modulus(arg):
	'''
	Find the modulus of the input as a function not an operator.
	'''

	# find the arguments of the function and cut off
	# everything else
	# sin(mod(5, 2)) ← the last parenthesis
	proto_inner = find_match(arg)

	# separate the arguments based on commas that are not
	# within more than one set of parentheses
	mod_args = separate(proto_inner[0][1:-1])

	if len(mod_args) != 2:
		raise CalculatorError(
			"mod takes exactly two arguments %s were"
			" given." % len(mod_args))

	# evaluate the arguments individually into a form that fmod
	# can accept
	inner1 = float(simplify(mod_args[0]))
	inner2 = float(simplify(mod_args[1]))

	# do the actual modulation
	result = math.fmod(inner1, inner2)

	# add on anything that was was cut off the end when finding
	# the arguments
	# sin(mod(5, 2)) ← the last parenthesis
	return(str(result) + proto_inner[1])


def abs_value(input):
	'''
	Break up a expression based on where pipes are and return the
	the absolute value of what is in them. Only does the first inner
	most layer, does not necessarily output a single number.

	>>> abs_value("|3-|2-4||")
	'|3-2.0|'

	>>> abs_value("|1-2|+|3-1|")
	'1.0+|3-1|'
	'''

	parts = input.split("|")
	
	if len(parts) % 2 == 0:
		raise CalculatorError(
			"There must be an even number of pipes in an "
			"absolute value expression.")

	for i in range(len(parts)):
		if parts[i].startswith(("+", "*", "^", "/")) or\
			parts[i].endswith(("+", "*", "^", "/", "-")) or\
			not parts[i]:
			pass

		else:
			result = math.fabs(float(simplify(parts[i])))
			last = ""
			iter_last = []
			next = ""
			iter_next = []

			if i > 0:
				last = parts[i - 1]
				if i > 1:
					iter_last = parts[:i - 1]

			if i < len(parts) - 1:
				next = parts[i + 1]
				if i < len(parts) - 2:
					iter_next = parts[i + 2:]

			result = last + str(result) + next
			result = "|".join(iter_last + [result] + iter_next)

			return(result)


def comma(left, right):
	'''
	Concatenate numbers separated by commas.
	'''

	if "." not in right:
		if left.isdigit() and len(right) == 3 and\
			right.isdigit():
			return(left + right)
		else:
			raise CalculatorError(
				"Commas used inappropriately.")
	else:
		parts = right.split(".")
		if left.isdigit() and len(parts[0]) == 3 and\
			(parts[1].isdigit() or parts[1] == ""):
			return(left + right)
		else:
			raise CalculatorError(
				"Commas used inappropriately.")


# main func
def simplify(s):
	'''
	Simplify an expression.

	This is the main body of the program.
	'''

	global degree_mode

	# iterates over all the operations
	for key, value in regular_expr.items():

		# solution to scientific notation being mistaken
		# for the constant e
		if key == "const_comp":
			try:
				s = float_to_str(s)
			except (ValueError, TypeError):
				pass

		# checks for the operation
		m = value.search(s)

		# continues until all instances of an
		# operation have been dealt with
		while m is not None:

			if key == "command_comp":
				# non-math commands

				# display history
				if m.group(1) is not None:
					print(history[-1 * hist_len:])

				# exit the program
				elif m.group(2) is not None:
					sys.exit()

				# set degree mode on for the session
				elif m.group(3) is not None:
					switch_degree_mode(2)

				# set degree mode off for the session
				elif m.group(4) is not None:
					switch_degree_mode(0)

				else:
					raise CalculatorError(
						"Command must be 'history',"
						" 'exit', 'quit', 'degree mode', or 'radian "
						"mode'.")

				return(None)

			elif key == "implicet_mult_pre_const_comp":

				result = m.group(1) + "*"

			elif key == "const_comp":

				result = constant(m.group(1))

			elif key == "implicet_mult_comp":

				result = m.group(1) + "*"

			elif key == "graph_comp":

				graph_out = graph_function(m.group(1))
				return(None)

			elif key == "alg_comp":

				result = solve_equations(m.group(1))

			elif key == "eval_comp":

				result = evaluate(m.group(1), m.group(2))

			elif key == "der_comp":

				if m.group(3) is not None:
					result = find_derivative(
						m.group(1), m.group(2),
						var = m.group(3)[-1])
				else:
					result = find_derivative(m.group(1), m.group(2))

			elif key == "int_comp":

				result = integrate_function(
					m.group(1), m.group(2),
					m.group(3), m.group(4))

			elif key in ("comb_comp", "choose_comp"):

				if key == "comb_comp":
					result = combinations_and_permutations(
						"func",
						m.group(1),
						m.group(2))
				elif key == "choose_comp":
					result = combinations_and_permutations(
						"choose",
						m.group(2),
						m.group(1),
						m.group(3))

			elif key == "ave_comp":

				result = statistics_functions(m.group(1), m.group(2))

			elif key == "trig_comp":

				result = single_argument(m.group(1), m.group(2))

			elif key in ("gamma_comp", "fact_comp"):

				if key == "gamma_comp":
					result = gamma(m.group(1))

				elif key == "fact_comp":
					result = factorial(m.group(1))

				else:
					raise CalculatorError(
						"How could this possibly "
						"happen? It just tested if key was "
						"'gamma_comp' or 'fact_comp'.")

			elif key == "log_comp":

				result = logarithm(m.group(1) or m.group(2))

			elif key == "abs_comp":

				result = abs_value(s)

			elif key == "paren_comp":

				# recursively evaluates the innermost parentheses
				result = simplify(m.group(1))

			elif key == "comma_comp":

				# just concatenates whats on either side
				# of the parentheses unless its separating
				# arguments of a function

				result = comma(m.group(1), m.group(2))

			elif key == "exp_comp":

				result = float(m.group(1)) ** float(m.group(3))

			elif key in ("mod_comp", "mod2_comp"):

				# modulus written as both mod(x, y) and x % y
				# where x is the dividend and y is the divisor
				# if written as x % y it will only take numbers
				# for arguments. In order of operations modulus comes
				# after exponents and factorials, but before
				# multiplication and division

				if key == "mod2_comp":

					result = modulus(m.group(1))

				elif key == "mod_comp":

					# the x % y format
					result = math.fmod(
						float(m.group(1)),
						float(m.group(2)))

				else:
					raise CalculatorError(
						"How could this possibly "
						"happen? It just tested if i was 'mod2_comp' "
						"or 'mod_comp'.")

			elif key == "per_comp":

				# percentage signs act just like dividing by 100
				result = "/100"

			elif key == "mult_comp":

				# multiplication and division

				if m.group(2) == "*":

					result = float(m.group(1)) * float(m.group(3))

				elif m.group(2) in ("/", "÷"):

					result = float(m.group(1)) / float(m.group(3))

				else:
					raise CalculatorError(
						"mult_comp must match '*', "
						"'//', or '÷'.")

			elif key == "add_comp":

				# addition and subtraction

				if m.group(2) == "+":

					result = math.fsum((
						float(m.group(1)),
						float(m.group(3))))

				elif m.group(2) == "-":

					result = float(m.group(1)) - float(m.group(3))

				else:
					raise CalculatorError(
						"add_comp must match '+' or '-'.")

			else:
				raise CalculatorError(
					"Function: '%s' not implemented." % key)

			if key not in (
				"command_comp", "const_comp",
				"alg_comp", "eval_comp", "der_comp"):

				# this is a fix for python returning
				# answers in scientific notation which since
				# it has e it mistakes the constant e
				try:
					result = float_to_str(result)
				except (ValueError, TypeError):
					pass

			# replace the text matched by i (the regular expression)
			# with the result of the mathematical expression
			s = sub(value, str(result), s, count = 1)

			m = value.search(s)
	try:
		s = float_to_str(s)
	except (ValueError, TypeError):
		pass
	return(s)


# pre and post processing for console
def ask(s = None):
	'''
	Ask the user what expression they want to simplify
	and do pre and post processing when using the console.
	'''

	global ans
	if s is None:

		# get input from the user
		s = input("input an expression: ")

		# add the user input to the history
		history.append(s)

		# save history to file
		save_info(history = history)

	try:
		# evaluate the expression
		out = simplify(s)
	except CalculatorError as e:
		print(s, " failed: ", e)
		logging.error(s + " failed: " + str(e))

	else:
		# save output to be used by the user
		ans = out

		# display the output
		if out is not None:
			print(s + " = " + out)
		print("")

		# save answer to file to be used next session
		save_info(ans = ans)


def key_pressed(event):
	'''
	Handle keys pressed in the GUI.
	'''

	global up_hist

	try:
		code = event.keycode
	except AttributeError:
		code = event

	if code in key_binds[os.name]:
		key = key_binds[os.name][code]
	else:
		key = None
	# print(code)

	if key == "enter":
		get_input()

	# go backwards in the history when the up arrow is pressed
	if key == "up":
		up_hist += 1
		input_widget.delete(0, "end")
		input_widget.insert(0, history[-1 * up_hist])

	# go forwards in the history when the down arrow is pressed
	if key == "down":
		if up_hist > 1:
			up_hist -= 1
			input_widget.delete(0, "end")
			input_widget.insert(0, history[-1 * up_hist])

	# if you are not navigating the history stop keeping
	# track of where you are
	if key not in ("up", "down"):
		up_hist = 0


def input_backspace():
	'''
	Delete the last character in the entry widget.
	'''

	global input_widget

	# delete the last character in the input widget
	a = input_widget.get()
	input_widget.delete(0, "end")
	input_widget.insert(0, a[:-1])


def get_input(s = None):
	'''
	Get user input from the entry widget.
	'''

	global ans, mess

	if s is None:
		pre_s = input_widget.get()
		if pre_s != "":
			s = pre_s

	if s == "":
		if os.name == "posix":
			exit()
		elif os.name == "nt":
			sys.exit()

	if s is not None:

		# add the user input to the history
		history.append(s)

		# save history to file
		save_info(history = history)

		try:
			out = simplify(s)
		except CalculatorError as e:
			mess.set(s + " failed: " + str(e))
			logging.error(s + " failed: " + str(e))
		else:
			# save output to be used by the user
			ans = out

			# display the output
			if out is not None:
				mess.set(s + " = " + out)

			# save answer to file to be used next session
			save_info(ans = ans)

		# clear the input box
		input_widget.delete(0, "end")


def switch_trig():
	'''
	Put the trig function buttons on the GUI.
	'''

	# remove the buttons for the hyperbolic functions, misc functions,
	# and stats functions
	for i in range(12):
		hyperbolic_func_buttons[i].grid_forget()
		try:
			misc_func_buttons[i].grid_forget()
		except IndexError:
			pass
		try:
			stats_func_buttons[i].grid_forget()
		except IndexError:
			pass

	for i in range(12):
		trig_func_buttons[i].grid(row = i // 3 + 3, column = i % 3 + 8)


def switch_hyperbolic():
	'''
	Put the hyperbolic function buttons in the GUI.
	'''

	# remove the buttons for the trig functions, misc functions,
	# and stats functions
	for i in range(12):
		trig_func_buttons[i].grid_forget()
		try:
			misc_func_buttons[i].grid_forget()
		except IndexError:
			pass
		try:
			stats_func_buttons[i].grid_forget()
		except IndexError:
			pass

	for i in range(12):
		hyperbolic_func_buttons[i].grid(
			row = i // 3 + 3,
			column = i % 3 + 8)


def switch_misc():
	'''
	Put the miscellaneous function buttons on the GUI.
	'''

	# remove the buttons for the trig functions, hyperbolic functions,
	# and stats functions
	for i in range(12):
		trig_func_buttons[i].grid_forget()
		hyperbolic_func_buttons[i].grid_forget()
		try:
			stats_func_buttons[i].grid_forget()
		except IndexError:
			pass

	for i in range(10):
		misc_func_buttons[i].grid(row = i // 3 + 3, column = i % 3 + 8)


def switch_stats():
	'''
	Put the statistics function buttons in the GUI.
	'''

	# remove the buttons for the trig functions, misc functions,
	# and hyperbolic functions
	for i in range(12):
		trig_func_buttons[i].grid_forget()
		hyperbolic_func_buttons[i].grid_forget()
		try:
			misc_func_buttons[i].grid_forget()
		except IndexError:
			pass

	# mean median mode
	for i in range(6):
		stats_func_buttons[i].grid(
			row = i // 3 + 3,
			column = i % 3 + 8)


def format_default_screen(menubar):
	'''
	Put the buttons in place on the GUI.
	'''

	# 7 8 9
	digit_button[7].grid(row = 3, column = 0)
	digit_button[8].grid(row = 3, column = 1)
	digit_button[9].grid(row = 3, column = 2)

	# 4 5 6
	digit_button[4].grid(row = 4, column = 0)
	digit_button[5].grid(row = 4, column = 1)
	digit_button[6].grid(row = 4, column = 2)

	# 1 2 3
	digit_button[1].grid(row = 5, column = 0)
	digit_button[2].grid(row = 5, column = 1)
	digit_button[3].grid(row = 5, column = 2)

	# 0 .
	digit_button[0].grid(row = 6, column = 0)
	digit_button[10].grid(row = 6, column = 1)  # .

	digit_button[11].grid(row = 3, column = 4)  # +
	digit_button[12].grid(row = 3, column = 5)  # -
	digit_button[13].grid(row = 4, column = 4)  # *
	digit_button[14].grid(row = 4, column = 5)  # ÷
	digit_button[15].grid(row = 5, column = 4)  # ^
	digit_button[16].grid(row = 5, column = 5)  # !
	digit_button[17].grid(row = 6, column = 4)  # π
	digit_button[18].grid(row = 6, column = 5)  # e
	digit_button[19].grid(row = 3, column = 6)  # (
	digit_button[20].grid(row = 3, column = 7)  # )
	digit_button[21].grid(row = 4, column = 6)  # |
	digit_button[22].grid(row = 4, column = 7)  # ,
	digit_button[23].grid(row = 5, column = 6)  # ∫
	digit_button[24].grid(row = 5, column = 7)  # x

	equals_button.grid(row = 6, column = 2)  # =
	back_button.grid(row = 3, column = 12)  # backspace

	# the functions menubutton
	menubar.grid(row = 3, column = 11)

	switch_trig()


def switch_matrices():
	'''
	Create window for dealing with matrices.
	'''

	pass


def graph_win_key_press(event, index):
	'''
	Deal with key presses while editing the graph window.
	'''

	global g_bound_entry

	try:
		code = event.keycode
	except AttributeError:
		code = event

	if code in key_binds[os.name]:
		if key_binds[os.name][code] == "up" and index > 0:
			g_bound_entry[g_bound_names[index - 1]].focus()
		elif key_binds[os.name][code] == "down" and\
			index < len(g_bound_names) - 1:
			g_bound_entry[g_bound_names[index + 1]].focus()


def edit_graph_window():
	'''
	Change the graph window options.
	'''

	global g_bound_entry, g_bound_string

	root = tk.Toplevel()

	g_bound_entry = {}
	g_bound_string = {}
	g_bound_disp = {}
	for index, val in enumerate(g_bound_names):
		g_bound_entry[val] = tk.Entry(root)

		g_bound_entry[val].grid(row = index, column = 1)

		g_bound_string[val] = tk.StringVar()
		g_bound_string[val].set("%s = %s" % (val, win_bound[val]))

		g_bound_disp[val] = tk.Message(
			root,
			textvariable = g_bound_string[val],
			width = 100)

		g_bound_disp[val].grid(row = index, column = 0)

		g_bound_entry[val].bind(
			"<Key>",
			lambda event, i = index: graph_win_key_press(event, i))

	root.bind("<Return>", lambda a: change_graph_win_set())

	root.mainloop()


def tkask(s = None):
	'''
	Make a GUI for the program.
	'''

	global input_widget, mess
	global digit_button, equals_button, back_button
	global trig_func_buttons, inverse_trig_func_buttons
	global hyperbolic_func_buttons, inverse_hyperbolic_func_buttons
	global misc_func_buttons, stats_func_buttons

	root = tk.Tk()

	root.title("ReCalc")

	if os.name == "nt":
		root.iconbitmap(default = os.path.join(
			calc_path,
			"ReCalc_icon.ico"))

	mess = tk.StringVar()
	mess.set("Input an expression")

	# output text widget
	output_mess_widget = tk.Message(
		root,
		textvariable = mess,
		width = 200)
	output_mess_widget.grid(row = 0, column = 0, columnspan = 11)

	# input text widget
	input_widget = tk.Entry(root, width = 90)
	input_widget.grid(row = 1, column = 0, columnspan = 12)

	# list of basic buttons
	button_keys = list(range(10)) + [
		".", "+", "-", "*", "÷", "^", "!", "π", "e", "(", ")", "|",
		",", "∫", "x"]

	# creating of the basic buttons
	digit_button = list(tk.Button(
		root,
		text = str(i),
		command = lambda k = i: input_widget.insert("end", k),
		width = 5) for i in button_keys)

	# equals button
	equals_button = tk.Button(
		root,
		text = "=",
		command = get_input,
		width = 5,
		bg = "light blue")

	# backspace button
	back_button = tk.Button(
		root,
		text = "delete",
		command = input_backspace,
		width = 5)

	# list of trig functions
	trig_funcs = (
		"sin(", "cos(", "tan(",
		"sec(", "csc(", "cot(",
		"arcsin(", "arccos(", "arctan(",
		"arcsec(", "arccsc(", "arccot(")

	# creating of trig function buttons
	trig_func_buttons = list(tk.Button(
		root,
		text = i[:-1],
		command = lambda k = i: input_widget.insert("end", k),
		width = 5) for i in trig_funcs)

	# list of hyperbolic functions
	hyperbolic_funcs = (
		"sinh(", "cosh(", "tanh(",
		"sech(", "csch(", "coth(",
		"arcsinh(", "arccosh(", "arctanh(",
		"arcsech(", "arccsch(", "arccoth(")

	# creation of hyperbolic function buttons
	hyperbolic_func_buttons = list(tk.Button(
		root,
		text = i[:-1],
		command = lambda k = i: input_widget.insert("end", k),
		width = 5) for i in hyperbolic_funcs)

	# list of misc fuctions
	misc_funcs = (
		"log(", "ln(", "Γ(", "abs(", "ceil(", "floor(",
		"erf(", "mod(", "C(", "P(")

	# creation of misc function buttons
	misc_func_buttons = list(tk.Button(
		root,
		text = i[:-1],
		command = lambda k = i: input_widget.insert("end", k),
		width = 5) for i in misc_funcs)

	# list of stats functions
	stats_funcs = (
		"mean(", "median(", "mode(", "stdev(", "max(", "min(")

	# creation of stats buttons
	stats_func_buttons = list(tk.Button(
		root,
		text = i[:-1],
		command = lambda k = i: input_widget.insert("end", k),
		width = 5) for i in stats_funcs)

	# more functions button
	menubar = tk.Menubutton(
		root,
		text = "Functions",
		relief = "raised")
	dropdown = tk.Menu(menubar, tearoff = 0)
	dropdown.add_command(
		label = "Trig Functions",
		command = switch_trig)
	dropdown.add_command(
		label = "Hyperbolic Functions",
		command = switch_hyperbolic)
	dropdown.add_command(
		label = "Misc Functions",
		command = switch_misc)
	dropdown.add_command(
		label = "Stats Functions",
		command = switch_stats)

	menubar.config(menu = dropdown)

	# menu at top right of screen
	options_menu = tk.Menu(root)
	list_options = tk.Menu(options_menu, tearoff = 0)

	list_options.add_command(
		label = "Radian Mode",
		command = lambda: switch_degree_mode("radian"))

	list_options.add_command(
		label = "Degree Mode",
		command = lambda: switch_degree_mode("degree"))

	list_options.add_command(
		label = "Polar Mode",
		command = lambda: switch_polar_mode("polar"))

	list_options.add_command(
		label = "Cartesian Mode",
		command = lambda: switch_polar_mode("Cartesian"))

	list_options.add_command(
		label = "Change length of history print back",
		command = change_hist_len_win)

	list_options.add_command(
		label = "Change der approx",
		command = change_der_approx_win)

	options_menu.add_cascade(label = "Options", menu = list_options)

	# list of other things the calculator can do
	matrices_plus = tk.Menu(options_menu, tearoff = 0)
	matrices_plus.add_command(
		label = "Matrices",
		command = switch_matrices)
	matrices_plus.add_command(
		label = "Window",
		command = edit_graph_window)
	options_menu.add_cascade(label = "Things", menu = matrices_plus)

	root.config(menu = options_menu)

	format_default_screen(menubar)

	root.bind("<Key>", key_pressed)
	
	get_input(s = s)

	root.mainloop()


def main():
	'''
	Handle command line arguments and decide whether or not to use
	the GUI.
	'''

	# default startup
	if len(sys.argv) == 1:
		if "tkinter" in sys.modules and use_gui:
			tkask()
		else:
			while True:
				ask()

	else:  # handling command line arguments and startup
		if "tkinter" in sys.modules and use_gui:
			tkask(" ".join(sys.argv[1:]))
		else:
			ask(" ".join(sys.argv[1:]))
			while True:
				ask()


if __name__ == "__main__":
	main()
